#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
#
# Code based on ``usb3_pipe``.
# SPDX-License-Identifier: BSD-3-Clause
""" Scrambling and descrambling for USB3. """

import operator
import functools

from amaranth import *

from .coding   import COM, stream_word_matches_symbol
from ...stream import USBRawSuperSpeedStream


#
# Scrambling modules.
# See [USB3.2r1: Appendix B].
#

class ScramblerLFSR(Elaboratable):
    """ Scrambler LFSR.

    Linear feedback shift register used for USB3 scrambling.
    Polynomial: X^16 + X^5 + X^4 + X^3 + 1

    See [USB3.2: Appendix B]

    Attributes
    ----------
    clear: Signal(), input
        Strobe; when high, resets the LFSR to its initial value.
    advance: Signal(), input
        Strobe; when high, the LFSR advances on each clock cycle.
    value: Signal(32), output
        The current value of the LFSR.

    Parameters
    ----------
    initial_value: 32-bit int, optional
        The initial value for the LFSR. Optional; defaults to all 1's, per the USB3 spec.
    """
    def __init__(self, initial_value=0xffff):
        self._initial_value = initial_value

        #
        # I/O port
        #
        self.clear   = Signal()
        self.advance = Signal()
        self.value   = Signal(32)


    def elaborate(self, platform):
        m = Module()

        next_value       = Signal(16)
        current_value    = Signal(16, init=self._initial_value)

        def xor_bits(*indices):
            bits = (current_value[i] for i in indices)
            return functools.reduce(operator.__xor__, bits)


        # Compute the next value in our internal LFSR state...
        m.d.comb += next_value.eq(Cat(
            xor_bits(0, 6, 8, 10),               # 0
            xor_bits(1, 7, 9, 11),               # 1
            xor_bits(2, 8, 10, 12),              # 2
            xor_bits(3, 6, 8, 9, 10, 11, 13),    # 3
            xor_bits(4, 6, 7, 8, 9, 11, 12, 14), # 4
            xor_bits(5, 6, 7, 9, 12, 13, 15),    # 5
            xor_bits(0, 6, 7, 8, 10, 13, 14),    # 6
            xor_bits(1, 7, 8, 9, 11, 14, 15),    # 7
            xor_bits(0, 2, 8, 9, 10, 12, 15),    # 8
            xor_bits(1, 3, 9, 10, 11, 13),       # 9
            xor_bits(0, 2, 4, 10, 11, 12, 14),   # 10
            xor_bits(1, 3, 5, 11, 12, 13, 15),   # 11
            xor_bits(2, 4, 6, 12, 13, 14),       # 12
            xor_bits(3, 5, 7, 13, 14, 15),       # 13
            xor_bits(4, 6, 8, 14, 15),           # 14
            xor_bits(5, 7, 9, 15)                # 15
        ))

        # Compute the LFSR's current output.
        m.d.comb += self.value.eq(Cat(
            current_value[15],
            current_value[14],
            current_value[13],
            current_value[12],
            current_value[11],
            current_value[10],
            current_value[9],
            current_value[8],
            current_value[7],
            current_value[6],
            current_value[5],
            xor_bits(4,  15),
            xor_bits(3,  14, 15),
            xor_bits(2,  13, 14, 15),
            xor_bits(1,  12, 13, 14),
            xor_bits(0,  11, 12, 13),
            xor_bits(10, 11, 12, 15),
            xor_bits(9,  10, 11, 14),
            xor_bits(8,  9 , 10, 13),
            xor_bits(7,  8 , 9,  12),
            xor_bits(6,  7 , 8,  11),
            xor_bits(5,  6 , 7,  10),
            xor_bits(4,  5 , 6,  9,  15),
            xor_bits(3,  4 , 5,  8,  14),
            xor_bits(2,  3 , 4,  7,  13, 15),
            xor_bits(1,  2 , 3,  6,  12, 14),
            xor_bits(0,  1 , 2,  5,  11, 13, 15),
            xor_bits(0,  1 , 4,  10, 12, 14),
            xor_bits(0,  3 , 9,  11, 13),
            xor_bits(2,  8 , 10, 12),
            xor_bits(1,  7 , 9,  11),
            xor_bits(0,  6 , 8,  10)
        ))

        # If we have a reset, clear our LFSR.
        with m.If(self.clear):
            m.d.ss += current_value.eq(self._initial_value)

        # Otherwise, advance when desired.
        with m.Elif(self.advance):
            m.d.ss += current_value.eq(next_value)

        return m


class Scrambler(Elaboratable):
    """ USB3-compliant data scrambler.

    Scrambles the transmitted data stream to reduce EMI.

    Attributes
    ----------
    clear: Signal(), input
        Strobe; when high, resets the scrambler to the start of its sequence.
    enable: Signal(), input
        When high, data scrambling is enabled. When low, data is passed through without scrambling.
    sink: USBRawSuperSpeedStream(), input stream
        The stream containing data to be scrambled.
    sink: USBRawSuperSpeedStream(), output stream
        The stream containing data the scrambled output.

    Parameters
    ----------
    initial_value: 32-bit int, optional
        The initial value for the LFSR. Optional.
    """
    def __init__(self, initial_value=0x7dbd):
        self._initial_value = initial_value

        #
        # I/O port
        #
        self.clear  = Signal()
        self.enable = Signal()
        self.hold   = Signal()

        self.sink   = USBRawSuperSpeedStream()
        self.source = USBRawSuperSpeedStream()

        # Debug signaling.
        self.lfsr_state = Signal.like(self.source.data)


    def elaborate(self, platform):
        m = Module()

        sink   = self.sink
        source = self.source

        # Detect when we're sending a comma; which should reset our scrambling LFSR.
        comma_present = stream_word_matches_symbol(sink, 0, symbol=COM)

        # Create our inner LFSR, which should advance whenever our input streams do.
        m.submodules.lfsr = lfsr = ScramblerLFSR(initial_value=self._initial_value)
        m.d.comb += [
            lfsr.clear    .eq(self.clear | comma_present),
            lfsr.advance  .eq(sink.valid & source.ready & ~self.hold)
        ]

        # Pass through non-scrambled signals directly.
        m.d.comb += [
            source.ctrl   .eq(sink.ctrl),
            source.valid  .eq(sink.valid),
            sink.ready    .eq(source.ready)
        ]


        # If we have any non-control words, scramble them by overriding our data assignment above
        # with the relevant data word XOR'd with our LFSR value. Note that control words are -never-
        # scrambled, per [USB3.2: Appendix B]
        for i in range(4):
            is_data_code = ~sink.ctrl[i]
            lfsr_word    = lfsr.value.word_select(i, 8)

            with m.If(self.enable & is_data_code):
                m.d.comb += source.data.word_select(i, 8).eq(sink.data.word_select(i, 8) ^ lfsr_word)
            with m.Else():
                m.d.comb += source.data.word_select(i, 8).eq(sink.data.word_select(i, 8))


        # Connect up our debug outputs.
        m.d.comb += [
            self.lfsr_state.eq(lfsr.value)
        ]


        return m



class Descrambler(Scrambler):
    """ USB3-compliant data descrambler.

    This module descrambles the received data stream. K-codes are not affected.
    This module automatically resets itself whenever a COM alignment character is seen.

    Attributes
    ----------
    enable: Signal(), input
        When high, data scrambling is enabled. When low, data is passed through without scrambling.
    sink: USBRawSuperSpeedStream(), input stream
        The stream containing data to be descrambled.
    source: USBRawSuperSpeedStream(), output stream
        The stream containing data the descrambled output.

    Parameters
    ----------
    initial_value: 32-bit int, optional
        The initial value for the LFSR. Optional.

    """
    def __init__(self, initial_value=0xffff):
        self._initial_value = initial_value
        super().__init__(initial_value=initial_value)



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test           import LunaGatewareTestCase, usb_domain_test_case

from amaranth import Record
from luna.gateware.usb.usb2.packet import USBDataPacketDeserializer, USBDataPacketGenerator, USBDataPacketReceiver
from luna.gateware.usb.usb2.packet import USBHandshakeDetector, USBHandshakeGenerator
from luna.gateware.usb.usb2.packet import InterpacketTimerInterface, USBInterpacketTimer, USBTokenDetector

class USBPacketizerTest(LunaGatewareTestCase):
    SYNC_CLOCK_FREQUENCY = None
    USB_CLOCK_FREQUENCY  = 60e6

    def instantiate_dut(self, extra_arguments=None):
        self.utmi = Record([
            ("rx_data",   8),
            ("rx_active", 1),
            ("rx_valid",  1)
        ])

        # If we don't have explicit extra arguments, use the base class's.
        if extra_arguments is None:
            extra_arguments = self.FRAGMENT_ARGUMENTS

        return self.FRAGMENT_UNDER_TEST(utmi=self.utmi, **extra_arguments)

    def provide_byte(self, byte):
        """ Provides a given byte on the UTMI receive data for one cycle. """
        yield self.utmi.rx_data.eq(byte)
        yield


    def start_packet(self, *, set_rx_valid=True):
        """ Starts a UTMI packet receive. """
        yield self.utmi.rx_active.eq(1)

        if set_rx_valid:
            yield self.utmi.rx_valid.eq(1)

        yield


    def end_packet(self):
        """ Starts a UTMI packet receive. """
        yield self.utmi.rx_active.eq(0)
        yield self.utmi.rx_valid.eq(0)
        yield


    def provide_packet(self, *octets, cycle_after=True):
        """ Provides an entire packet transaction at once; for convenience. """
        yield from self.start_packet()
        for b in octets:
            yield from self.provide_byte(b)
        yield from self.end_packet()

        if cycle_after:
            yield


class USBTokenDetectorTest(USBPacketizerTest):
    FRAGMENT_UNDER_TEST = USBTokenDetector

    @usb_domain_test_case
    def test_valid_token(self):
        dut = self.dut

        # Assume our device is at address 0x3a.
        yield dut.address.eq(0x3a)

        # When idle, we should have no new-packet events.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.interface.new_frame), 0)
        self.assertEqual((yield dut.interface.new_token), 0)

        # From: https://usb.org/sites/default/files/crcdes.pdf
        # out to 0x3a, endpoint 0xa => 0xE1 5C BC
        yield from self.provide_packet(0b11100001, 0b00111010, 0b00111101)

        # Validate that we just finished a token.
        self.assertEqual((yield dut.interface.new_token), 1)
        self.assertEqual((yield dut.interface.new_frame), 0)

        # Validate that we got the expected PID.
        self.assertEqual((yield dut.interface.pid), 0b0001)

        # Validate that we got the expected address / endpoint.
        self.assertEqual((yield dut.interface.address),  0x3a)
        self.assertEqual((yield dut.interface.endpoint), 0xa )

        # Ensure that our strobe returns to 0, afterwards.
        yield
        self.assertEqual((yield dut.interface.new_token), 0)


    @usb_domain_test_case
    def test_valid_start_of_frame(self):
        dut = self.dut
        yield from self.provide_packet(0b10100101, 0b00111010, 0b00111101)

        # Validate that we just finished a token.
        self.assertEqual((yield dut.interface.new_token), 0)
        self.assertEqual((yield dut.interface.new_frame), 1)

        # Validate that we got the expected address / endpoint.
        self.assertEqual((yield dut.interface.frame), 0x53a)


    @usb_domain_test_case
    def test_token_to_other_device(self):
        dut = self.dut

        # Assume our device is at 0x1f.
        yield dut.address.eq(0x1f)

        # From: https://usb.org/sites/default/files/crcdes.pdf
        # out to 0x3a, endpoint 0xa => 0xE1 5C BC
        yield from self.provide_packet(0b11100001, 0b00111010, 0b00111101)

        # Validate that we did not count this as a token received,
        # as it wasn't for us.
        self.assertEqual((yield dut.interface.new_token), 0)


class USBHandshakeDetectorTest(USBPacketizerTest):
    FRAGMENT_UNDER_TEST = USBHandshakeDetector

    @usb_domain_test_case
    def test_ack(self):
        yield from self.provide_packet(0b11010010)
        self.assertEqual((yield self.dut.detected.ack), 1)

    @usb_domain_test_case
    def test_nak(self):
        yield from self.provide_packet(0b01011010)
        self.assertEqual((yield self.dut.detected.nak), 1)

    @usb_domain_test_case
    def test_stall(self):
        yield from self.provide_packet(0b00011110)
        self.assertEqual((yield self.dut.detected.stall), 1)

    @usb_domain_test_case
    def test_nyet(self):
        yield from self.provide_packet(0b10010110)
        self.assertEqual((yield self.dut.detected.nyet), 1)

class USBDataPacketReceiverTest(USBPacketizerTest):
    FRAGMENT_UNDER_TEST = USBDataPacketReceiver
    FRAGMENT_ARGUMENTS  = {'standalone': True}

    @usb_domain_test_case
    def test_data_receive(self):
        dut = self.dut
        stream = self.dut.stream


        #    0xC3,                                           # PID: Data
        #    0x00, 0x05, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, # DATA
        #    0xEB, 0xBC                                      # CRC

        # Before we send anything, our stream should be inactive.
        self.assertEqual((yield stream.valid), 0)

        # After sending our PID, we shouldn't see our stream be active.
        yield from self.start_packet()
        yield from self.provide_byte(0xc3)
        self.assertEqual((yield stream.valid), 0)

        # The stream shouldn't go valid for our first two data bytes, either.
        for b in [0x00, 0x05]:
            yield from self.provide_byte(b)
            self.assertEqual((yield stream.valid), 0)

        # Check that our active PID was successfully captured, and our current on
        # hasn't yet been updated.
        self.assertEqual((yield self.dut.active_pid), 3)
        self.assertEqual((yield self.dut.packet_id),  0)

        # The third byte should finally trigger our stream output...
        yield from self.provide_byte(0x08)
        self.assertEqual((yield stream.valid), 1)

        # ... and we should see the first byte on our stream.
        self.assertEqual((yield stream.next),       1)
        self.assertEqual((yield stream.payload), 0x00)

        # If we pause RxValid, we nothing should advance.
        yield self.utmi.rx_valid.eq(0)

        yield from self.provide_byte(0x08)
        self.assertEqual((yield stream.next),       0)
        self.assertEqual((yield stream.payload), 0x00)

        # Resuming should continue our advance...
        yield self.utmi.rx_valid.eq(1)

        # ... and we should process the remainder of the input
        for b in [0x00, 0x00, 0x00, 0x00, 0x00, 0xEB, 0xBC]:
            yield from self.provide_byte(b)

        # ... remaining two bytes behind.
        self.assertEqual((yield stream.next),       1)
        self.assertEqual((yield stream.payload), 0x00)


        # When we stop our packet, we should see our stream stop as well.
        # The last two bytes, our CRC, shouldn't be included.
        yield from self.end_packet()
        yield

        # And, since we sent a valid packet, we should see a pulse indicating the packet is valid.
        self.assertEqual((yield stream.valid),              0)
        self.assertEqual((yield self.dut.packet_complete),  1)

        # After an inter-packet delay, we should see that we're ready to respond.
        yield from self.advance_cycles(9)
        self.assertEqual((yield self.dut.ready_for_response), 0)


    @usb_domain_test_case
    def test_zlp(self):
        dut = self.dut
        stream = self.dut.stream

        #    0x4B        # DATA1
        #    0x00, 0x00  # CRC

        # Send our data PID.
        yield from self.start_packet()
        yield from self.provide_byte(0x4B)

        # Send our CRC.
        for b in [0x00, 0x00]:
            yield from self.provide_byte(b)
            self.assertEqual((yield stream.valid), 0)

        yield from self.end_packet()
        yield

        self.assertEqual((yield self.dut.packet_complete),  1)


class USBDataPacketDeserializerTest(USBPacketizerTest):
    FRAGMENT_UNDER_TEST = USBDataPacketDeserializer

    def instantiate_dut(self):
        return super().instantiate_dut(extra_arguments={'create_crc_generator': True})


    @usb_domain_test_case
    def test_packet_rx(self):
        yield from self.provide_packet(
            0b11000011,                                     # PID
            0b00100011, 0b01000101, 0b01100111, 0b10001001, # DATA
            0b00001110, 0b00011100                          # CRC
        )

        # Ensure we've gotten a new packet.
        self.assertEqual((yield self.dut.new_packet), 1, "packet not recognized")
        self.assertEqual((yield self.dut.length),     4)
        self.assertEqual((yield self.dut.packet[0]),  0b00100011)
        self.assertEqual((yield self.dut.packet[1]),  0b01000101)
        self.assertEqual((yield self.dut.packet[2]),  0b01100111)
        self.assertEqual((yield self.dut.packet[3]),  0b10001001)


    @usb_domain_test_case
    def test_captured_usb_sample(self):
        yield from self.provide_packet(
            0xC3,                                           # PID: Data
            0x00, 0x05, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, # DATA
            0xEB, 0xBC                                      # CRC
        )

        # Ensure we've gotten a new packet.
        self.assertEqual((yield self.dut.new_packet), 1, "packet not recognized")

    @usb_domain_test_case
    def test_invalid_rx(self):
        yield from self.provide_packet(
            0b11000011,                                     # PID
            0b11111111, 0b11111111, 0b11111111, 0b11111111, # DATA
            0b00011100, 0b00001110                          # CRC
        )

        # Ensure we've gotten a new packet.
        self.assertEqual((yield self.dut.new_packet), 0, 'accepted invalid CRC!')

class USBDataPacketGeneratorTest(LunaGatewareTestCase):
    SYNC_CLOCK_FREQUENCY = None
    USB_CLOCK_FREQUENCY = 60e6

    FRAGMENT_UNDER_TEST = USBDataPacketGenerator
    FRAGMENT_ARGUMENTS  = {'standalone': True}

    def initialize_signals(self):
        # Model our PHY is always accepting data, by default.
        yield self.dut.tx.ready.eq(1)


    @usb_domain_test_case
    def test_simple_data_generation(self):
        dut    = self.dut
        stream = self.dut.stream
        tx     = self.dut.tx

        # We'll request that a simple USB packet be sent. We expect the following data:
        #    0xC3,                                           # PID: Data
        #    0x00, 0x05, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, # DATA
        #    0xEB, 0xBC                                      # CRC

        # Before we send anything, we shouldn't be transmitting.
        self.assertEqual((yield dut.tx.valid), 0)

        # Work with DATA0 packet IDs.
        yield dut.data_pid.eq(0)

        # Start sending our first byte.
        yield stream.first.eq(1)
        yield stream.valid.eq(1)
        yield stream.payload.eq(0x00)
        yield

        # Once our first byte has been provided, our transmission should
        # start (valid=1), and we should see our data PID.
        yield
        self.assertEqual((yield tx.valid), 1)
        self.assertEqual((yield tx.data), 0xc3)

        # We shouldn't consume any data, yet, as we're still
        # transmitting our PID.
        self.assertEqual((yield stream.ready), 0)

        # Drop our first value back to zero, as it should also work as a strobe.
        yield stream.first.eq(0)

        # One cycle later, we should see our first data byte, and our
        # stream should indicate that data was consumed.
        yield
        self.assertEqual((yield tx.data), 0x00)
        self.assertEqual((yield stream.ready), 1)

        # Provide the remainder of our data, and make sure that our
        # output value mirrors it.
        for datum in [0x05, 0x08, 0x00, 0x00, 0x00, 0x00]:
            yield stream.payload.eq(datum)
            yield
            self.assertEqual((yield tx.data), datum)

        # Finally, provide our last data value.
        yield stream.payload.eq(0x00)
        yield stream.last.eq(1)
        yield

        # Drop our stream-valid to zero after the last stream byte.
        yield stream.valid.eq(0)

        # We should now see that we're no longer consuming data...
        yield
        self.assertEqual((yield stream.ready), 0)

        # ... but the transmission is still valid; and now presenting our CRC...
        self.assertEqual((yield tx.valid), 1)
        self.assertEqual((yield tx.data),  0xeb)

        # ... which is two-bytes long.
        yield
        self.assertEqual((yield tx.valid), 1)
        self.assertEqual((yield tx.data),  0xbc)

        # Once our CRC is completed, our transmission should stop.
        yield
        self.assertEqual((yield tx.valid), 0)


    @usb_domain_test_case
    def test_single_byte(self):
        stream = self.dut.stream

        # Request single byte.
        yield stream.first.eq(1)
        yield stream.last.eq(1)
        yield stream.valid.eq(1)
        yield stream.payload.eq(0xAB)
        yield from self.wait_until(stream.ready)

        # Drop our last back to zero, immediately.
        yield stream.last.eq(0)
        yield stream.first.eq(0)
        yield stream.valid.eq(0)

        yield from self.advance_cycles(10)


    @usb_domain_test_case
    def test_zlp_generation(self):
        stream = self.dut.stream
        tx     = self.dut.tx

        # Request a ZLP.
        yield stream.first.eq(0)
        yield stream.last.eq(1)
        yield stream.valid.eq(1)
        yield

        # Drop our last back to zero, immediately.
        yield stream.last.eq(0)

        # Once our first byte has been provided, our transmission should
        # start (valid=1), and we should see our data PID.
        yield
        self.assertEqual((yield tx.valid), 1)
        self.assertEqual((yield tx.data), 0xc3)

        # Drop our stream-valid to zero after the last stream byte.
        yield stream.valid.eq(0)

        # We should now see that we're no longer consuming data...
        yield
        self.assertEqual((yield stream.ready), 0)

        # ... but the transmission is still valid; and now presenting our CRC...
        self.assertEqual((yield tx.valid), 1)
        self.assertEqual((yield tx.data),  0x0)

        # ... which is two-bytes long.
        yield
        self.assertEqual((yield tx.valid), 1)
        self.assertEqual((yield tx.data),  0x0)


class USBHandshakeGeneratorTest(LunaGatewareTestCase):
    SYNC_CLOCK_FREQUENCY = None
    USB_CLOCK_FREQUENCY = 60e6

    FRAGMENT_UNDER_TEST  = USBHandshakeGenerator


    @usb_domain_test_case
    def test_ack_generation(self):
        dut = self.dut

        # Before we request anything, our data shouldn't be valid.
        self.assertEqual((yield dut.tx.valid), 0)

        # When we request an ACK...
        yield dut.issue_ack.eq(1)
        yield
        yield dut.issue_ack.eq(0)

        # ... we should see an ACK packet on our data lines...
        yield
        self.assertEqual((yield dut.tx.data), USBHandshakeGenerator._PACKET_ACK)

        # ... our transmit request should be valid.
        self.assertEqual((yield dut.tx.valid), 1)

        # It should remain valid...
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.tx.valid), 1)

        # ... until the UTMI transceiver marks it as accepted...
        yield dut.tx.ready.eq(1)
        yield

        # ... when our packet should be marked as invalid.
        yield
        self.assertEqual((yield dut.tx.valid), 0)


    @usb_domain_test_case
    def test_already_ready(self):
        dut = self.dut

        # Start off with our transmitter ready to receive.
        yield dut.tx.ready.eq(1)

        # When we request an ACK...
        yield dut.issue_ack.eq(1)
        yield
        yield dut.issue_ack.eq(0)

        # ... we should see an ACK packet on our data lines...
        yield
        self.assertEqual((yield dut.tx.data), USBHandshakeGenerator._PACKET_ACK)

        # ... our transmit request should be valid...
        self.assertEqual((yield dut.tx.valid), 1)

        # ... and then drop out of being valid after one cycle.
        yield
        self.assertEqual((yield dut.tx.valid), 0)


class USBInterpacketTimerTest(LunaGatewareTestCase):
    SYNC_CLOCK_FREQUENCY = None
    USB_CLOCK_FREQUENCY = 60e6

    def instantiate_dut(self):
        dut = USBInterpacketTimer()

        # Create our primary timer interface.
        self.interface = InterpacketTimerInterface()
        dut.add_interface(self.interface)

        return dut


    def initialize_signals(self):
        # Assume FS for our tests, unless overridden.
        yield self.dut.speed(USBSpeed.FULL)


    def test_resets_and_delays(self):
        yield from self.advance_cycles(4)
        interface = self.interface

        # Trigger a cycle reset.
        yield interface.start.eq(1)
        yield
        yield interface.start.eq(0)

        # We should start off with no timer outputs high.
        self.assertEqual((yield interface.tx_allowed), 0)
        self.assertEqual((yield interface.tx_timeout), 0)
        self.assertEqual((yield interface.rx_timeout), 0)

        # 10 cycles later, we should see our first timer output.
        yield from self.advance_cycles(10)
        self.assertEqual((yield interface.tx_allowed), 1)
        self.assertEqual((yield interface.tx_timeout), 0)
        self.assertEqual((yield interface.rx_timeout), 0)

        # 22 cycles later (32 total), we should see our second timer output.
        yield from self.advance_cycles(22)
        self.assertEqual((yield interface.tx_allowed), 0)
        self.assertEqual((yield interface.tx_timeout), 1)
        self.assertEqual((yield interface.rx_timeout), 0)

        # 58 cycles later (80 total), we should see our third timer output.
        yield from self.advance_cycles(22)
        self.assertEqual((yield interface.tx_allowed), 0)
        self.assertEqual((yield interface.tx_timeout), 0)
        self.assertEqual((yield interface.rx_timeout), 1)

        # Ensure that the timers don't go high again.
        for _ in range(32):
            self.assertEqual((yield self.rx_to_tx_min),     0)
            self.assertEqual((yield self.rx_to_tx_max),     0)
            self.assertEqual((yield self.tx_to_rx_timeout), 0)



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test.utils import LunaGatewareTestCase, sync_test_case

from luna.gateware.interface.psram import HyperBusPHY, HyperRAMInterface

class TestHyperRAMInterface(LunaGatewareTestCase):

    def instantiate_dut(self):
        # Create a record that recreates the layout of our RAM signals.
        self.ram_signals = HyperBusPHY()

        # Create our HyperRAM interface...
        return HyperRAMInterface(phy=self.ram_signals)


    def assert_clock_pulses(self, times=1):
        """ Function that asserts we get a specified number of clock pulses. """

        for _ in range(times):
            yield
            self.assertEqual((yield self.ram_signals.clk_en), 1)


    @sync_test_case
    def test_register_write(self):

        # Before we transact, CS should be de-asserted, and RWDS and DQ should be undriven.
        yield
        self.assertEqual((yield self.ram_signals.cs),      0)
        self.assertEqual((yield self.ram_signals.dq.e),    0)
        self.assertEqual((yield self.ram_signals.rwds.e),  0)

        yield from self.advance_cycles(10)
        self.assertEqual((yield self.ram_signals.cs),      0)

        # Request a register write to ID register 0.
        yield self.dut.perform_write  .eq(1)
        yield self.dut.register_space .eq(1)
        yield self.dut.address        .eq(0x00BBCCDD)
        yield self.dut.start_transfer .eq(1)
        yield self.dut.final_word     .eq(1)
        yield self.dut.write_data     .eq(0xBEEF)

        # Simulate the RAM requesting a extended latency.
        yield self.ram_signals.rwds.i .eq(1)
        yield

        # Ensure that upon requesting, CS goes high, and our clock starts low.
        yield
        self.assertEqual((yield self.ram_signals.cs),     1)
        self.assertEqual((yield self.ram_signals.clk_en), 0)

        # Drop our "start request" line somewhere during the transaction;
        # so we don't immediately go into the next transfer.
        yield self.dut.start_transfer.eq(0)

        # We should then move to shifting out our first command word,
        # which means we're driving DQ with the first word of our command.
        yield
        yield
        self.assertEqual((yield self.ram_signals.cs),       1)
        self.assertEqual((yield self.ram_signals.clk_en),   1)
        self.assertEqual((yield self.ram_signals.dq.e),     1)
        self.assertEqual((yield self.ram_signals.dq.o),  0x6017)

        # This should continue until we've shifted out a full command.
        yield
        self.assertEqual((yield self.ram_signals.dq.o),  0x799B)
        yield
        self.assertEqual((yield self.ram_signals.dq.o),  0x0005)

        # Check that we've been driving our output this whole time,
        # and haven't been driving RWDS.
        self.assertEqual((yield self.ram_signals.dq.e),    1)
        self.assertEqual((yield self.ram_signals.rwds.e),  0)
        yield

        # For a _register_ write, there shouldn't be latency period.
        # This means we should continue driving DQ...
        self.assertEqual((yield self.ram_signals.dq.e),    1)
        self.assertEqual((yield self.ram_signals.rwds.e),  0)
        self.assertEqual((yield self.ram_signals.dq.o),  0xBEEF)



    @sync_test_case
    def test_register_read(self):

        # Before we transact, CS should be de-asserted, and RWDS and DQ should be undriven.
        yield
        self.assertEqual((yield self.ram_signals.cs),      0)
        self.assertEqual((yield self.ram_signals.dq.e),   0)
        self.assertEqual((yield self.ram_signals.rwds.e), 0)

        yield from self.advance_cycles(10)
        self.assertEqual((yield self.ram_signals.cs),      0)

        # Request a register read of ID register 0.
        yield self.dut.perform_write  .eq(0)
        yield self.dut.register_space .eq(1)
        yield self.dut.address        .eq(0x00BBCCDD)
        yield self.dut.start_transfer .eq(1)
        yield self.dut.final_word     .eq(1)

        # Simulate the RAM requesting a extended latency.
        yield self.ram_signals.rwds.i .eq(1)
        yield

        # Ensure that upon requesting, CS goes high, and our clock starts low.
        yield
        self.assertEqual((yield self.ram_signals.cs),     1)
        self.assertEqual((yield self.ram_signals.clk_en), 0)

        # Drop our "start request" line somewhere during the transaction;
        # so we don't immediately go into the next transfer.
        yield self.dut.start_transfer.eq(0)

        # We should then move to shifting out our first command word,
        # which means we're driving DQ with the first word of our command.
        yield
        yield
        self.assertEqual((yield self.ram_signals.cs),         1)
        self.assertEqual((yield self.ram_signals.clk_en),     1)
        self.assertEqual((yield self.ram_signals.dq.e),       1)
        self.assertEqual((yield self.ram_signals.dq.o),  0xe017)

        # This should continue until we've shifted out a full command.
        yield
        self.assertEqual((yield self.ram_signals.dq.o),  0x799B)
        yield
        self.assertEqual((yield self.ram_signals.dq.o),  0x0005)

        # Check that we've been driving our output this whole time,
        # and haven't been driving RWDS.
        self.assertEqual((yield self.ram_signals.dq.e),    1)
        self.assertEqual((yield self.ram_signals.rwds.e),  0)

        # Once we finish scanning out the word, we should stop driving
        # the data lines, and should finish two latency periods before
        # sending any more data.
        yield
        self.assertEqual((yield self.ram_signals.dq.e),    0)
        self.assertEqual((yield self.ram_signals.rwds.e),  0)
        self.assertEqual((yield self.ram_signals.clk_en),  1)

        # By this point, the RAM will drive RWDS low.
        yield self.ram_signals.rwds.i.eq(0)

        # Ensure the clock still ticking...
        yield
        self.assertEqual((yield self.ram_signals.clk_en),    1)

        # ... and remains so for the remainder of the latency period.
        yield from self.assert_clock_pulses(14)

        # Now, shift in a pair of data words.
        yield self.ram_signals.dq.i.eq(0xCAFE)
        yield self.ram_signals.rwds.i.eq(0b10)
        yield

        # Once this finished, we should have a result on our data out.
        self.assertEqual((yield self.dut.read_data),      0xCAFE)
        self.assertEqual((yield self.dut.read_ready    ), 1)

        yield
        yield
        self.assertEqual((yield self.ram_signals.cs),     0)
        self.assertEqual((yield self.ram_signals.dq.e),   0)
        self.assertEqual((yield self.ram_signals.rwds.e), 0)

        # Ensure that our clock drops back to '0' during idle cycles.
        yield from self.advance_cycles(2)
        self.assertEqual((yield self.ram_signals.clk_en),  0)

        # TODO: test recovery time




#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test          import LunaGatewareTestCase, usb_domain_test_case

from luna.gateware.usb.usb2.transfer import USBInTransferManager

class USBInTransferManagerTest(LunaGatewareTestCase):
    FRAGMENT_UNDER_TEST = USBInTransferManager
    FRAGMENT_ARGUMENTS  = {"max_packet_size": 8}

    SYNC_CLOCK_FREQUENCY = None
    USB_CLOCK_FREQUENCY = 60e6

    def initialize_signals(self):

        # By default, pretend our transmitter is always accepting data...
        yield self.dut.packet_stream.ready.eq(1)

        # And pretend that our host is always tagreting our endpoint.
        yield self.dut.active.eq(1)
        yield self.dut.tokenizer.is_in.eq(1)


    @usb_domain_test_case
    def test_normal_transfer(self):
        dut = self.dut

        packet_stream   = dut.packet_stream
        transfer_stream = dut.transfer_stream

        # Before we do anything, we shouldn't have anything our output stream.
        self.assertEqual((yield packet_stream.valid), 0)

        # Our transfer stream should accept data until we fill up its buffers.
        self.assertEqual((yield transfer_stream.ready), 1)

        # Once we start sending data to our packetizer...
        yield transfer_stream.valid.eq(1)
        yield transfer_stream.payload.eq(0x11)
        yield

        # We still shouldn't see our packet stream start transmitting;
        # and we should still be accepting data.
        self.assertEqual((yield packet_stream.valid), 0)
        self.assertEqual((yield transfer_stream.ready), 1)

        # Once we see a full packet...
        for value in [0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            yield transfer_stream.payload.eq(value)
            yield
        yield transfer_stream.valid.eq(0)

        # ... we shouldn't see a transmit request until we receive an IN token.
        self.assertEqual((yield transfer_stream.ready), 1)
        yield from self.advance_cycles(5)
        self.assertEqual((yield packet_stream.valid), 0)

        # We -should-, however, keep filling our secondary buffer while waiting.
        yield transfer_stream.valid.eq(1)
        self.assertEqual((yield transfer_stream.ready), 1)
        for value in [0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF, 0x00]:
            yield transfer_stream.payload.eq(value)
            yield

        # Once we've filled up -both- buffers, our data should no longer be ready.
        yield
        self.assertEqual((yield transfer_stream.ready), 0)

        # Once we do see an IN token...
        yield from self.pulse(dut.tokenizer.ready_for_response)

        # ... we should start transmitting...
        self.assertEqual((yield packet_stream.valid), 1)

        # ... we should see the full packet be emitted...
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            self.assertEqual((yield packet_stream.payload), value)
            yield

        # ... and then the packet should end.
        self.assertEqual((yield packet_stream.valid), 0)

        # We should now be waiting for an ACK. While waiting, we still need
        # to keep the last packet; so we'll expect that we're not ready for data.
        self.assertEqual((yield transfer_stream.ready), 0)

        # If we receive anything other than an ACK...
        yield from self.pulse(dut.tokenizer.new_token)
        yield

        # ... we should see the same data transmitted again, with the same PID.
        yield from self.pulse(dut.tokenizer.ready_for_response, step_after=False)
        yield self.assertEqual((yield dut.data_pid), 0)

        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            self.assertEqual((yield packet_stream.payload), value)
            yield

        # If we do ACK...
        yield from self.pulse(dut.handshakes_in.ack)

        # ... we should see our DATA PID flip, and we should be ready to accept data again...
        yield self.assertEqual((yield dut.data_pid), 1)
        yield self.assertEqual((yield transfer_stream.ready), 1)

        #  ... and we should get our second packet.
        yield from self.pulse(dut.tokenizer.ready_for_response, step_after=True)
        for value in [0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF, 0x00]:
            self.assertEqual((yield packet_stream.payload), value)
            yield


    @usb_domain_test_case
    def test_nak_when_not_ready(self):
        dut = self.dut

        # We shouldn't initially be NAK'ing anything...
        self.assertEqual((yield dut.handshakes_out.nak), 0)

        # ... but if we get an IN token we're not ready for...
        yield from self.pulse(dut.tokenizer.ready_for_response, step_after=False)

        # ... we should see one cycle of NAK.
        self.assertEqual((yield dut.handshakes_out.nak), 1)
        yield
        self.assertEqual((yield dut.handshakes_out.nak), 0)


    @usb_domain_test_case
    def test_zlp_generation(self):
        dut = self.dut

        packet_stream   = dut.packet_stream
        transfer_stream = dut.transfer_stream

        # Simulate a case where we're generating ZLPs.
        yield dut.generate_zlps.eq(1)


        # If we're sent a full packet _without the transfer stream ending_...
        yield transfer_stream.valid.eq(1)
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            yield transfer_stream.payload.eq(value)
            yield
        yield transfer_stream.valid.eq(0)


        # ... we should receive that data packet without a ZLP.
        yield from self.pulse(dut.tokenizer.ready_for_response)
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            self.assertEqual((yield packet_stream.payload), value)
            yield
        self.assertEqual((yield dut.data_pid), 0)
        yield from self.pulse(dut.handshakes_in.ack)


        # If we send a full packet...
        yield transfer_stream.valid.eq(1)
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77]:
            yield transfer_stream.payload.eq(value)
            yield

        # ... that _ends_ our transfer...
        yield transfer_stream.payload.eq(0x88)
        yield transfer_stream.last.eq(1)
        yield

        yield transfer_stream.last.eq(0)
        yield transfer_stream.valid.eq(0)

        # ... we should emit the relevant data packet...
        yield from self.pulse(dut.tokenizer.ready_for_response)
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            self.assertEqual((yield packet_stream.payload), value)
            yield
        self.assertEqual((yield dut.data_pid), 1)
        yield from self.pulse(dut.handshakes_in.ack)

        # ... followed by a ZLP.
        yield from self.pulse(dut.tokenizer.ready_for_response, step_after=False)
        self.assertEqual((yield packet_stream.last), 1)
        self.assertEqual((yield dut.data_pid), 0)
        yield from self.pulse(dut.handshakes_in.ack)


        # Finally, if we're sent a short packet that ends our stream...
        yield transfer_stream.valid.eq(1)
        for value in [0xAA, 0xBB, 0xCC]:
            yield transfer_stream.payload.eq(value)
            yield
        yield transfer_stream.payload.eq(0xDD)
        yield transfer_stream.last.eq(1)

        yield
        yield transfer_stream.last.eq(0)
        yield transfer_stream.valid.eq(0)

        # ... we should emit the relevant short packet...
        yield from self.pulse(dut.tokenizer.ready_for_response)
        for value in [0xAA, 0xBB, 0xCC, 0xDD]:
            self.assertEqual((yield packet_stream.payload), value)
            yield
        yield from self.pulse(dut.handshakes_in.ack)
        self.assertEqual((yield dut.data_pid), 1)


        # ... and we shouldn't emit a ZLP; meaning we should be ready to receive new data.
        self.assertEqual((yield transfer_stream.ready), 1)


    @usb_domain_test_case
    def test_discard(self):
        dut = self.dut

        packet_stream   = dut.packet_stream
        transfer_stream = dut.transfer_stream

        # Before we do anything, we shouldn't have anything our output stream.
        self.assertEqual((yield packet_stream.valid), 0)

        # Our transfer stream should accept data until we fill up its buffers.
        self.assertEqual((yield transfer_stream.ready), 1)

        # We queue up two full packets.
        yield transfer_stream.valid.eq(1)
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            yield transfer_stream.payload.eq(value)
            yield

        for value in [0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF, 0x00]:
            yield transfer_stream.payload.eq(value)
            yield
        yield transfer_stream.valid.eq(0)

        # Once we do see an IN token...
        yield from self.pulse(dut.tokenizer.ready_for_response)

        # ... we should start transmitting...
        self.assertEqual((yield packet_stream.valid), 1)

        # ... and should see the full packet be emitted...
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            self.assertEqual((yield packet_stream.payload), value)
            yield

        # ... with DATA PID 0 ...
        self.assertEqual((yield dut.data_pid), 0)

        # ... and then the packet should end.
        self.assertEqual((yield packet_stream.valid), 0)

        # If we ACK the first packet...
        yield from self.pulse(dut.handshakes_in.ack)

        # ... we should be ready to accept data again.
        self.assertEqual((yield transfer_stream.ready), 1)
        yield from self.advance_cycles(5)

        # If we then discard the second packet...
        yield from self.pulse(dut.discard, step_after=False)

        # ... we shouldn't see a transmit request upon an in token.
        yield from self.pulse(dut.tokenizer.ready_for_response, step_after=False)
        yield from self.advance_cycles(5)
        self.assertEqual((yield packet_stream.valid), 0)

        # If we send another full packet...
        yield transfer_stream.valid.eq(1)
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            yield transfer_stream.payload.eq(value)
            yield
        yield transfer_stream.valid.eq(0)

        # ... and see an IN token...
        yield from self.pulse(dut.tokenizer.ready_for_response)

        # ... we should start transmitting...
        self.assertEqual((yield packet_stream.valid), 1)

        # ... add should see the full packet be emitted...
        for value in [0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]:
            self.assertEqual((yield packet_stream.payload), value)
            yield

        # ... with the correct DATA PID.
        self.assertEqual((yield dut.data_pid), 1)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Link Training and Status State Machine (LTSSM) gateware. """

#
# WARNING: This implementation is currently the minimum set of things that make
# the SerDes PHY work in some cases. This is not a complete design.
#

import math

from amaranth import *
from amaranth.lib.coding import Encoder
from amaranth.lib.cdc  import PulseSynchronizer



class LTSSMController(Elaboratable):
    """ Link Training and Status State Machine

    This state machine orchestrates USB bringup, link training, and power saving.
    It is implemented according to chaper 7.5 of the USB 3.2 specification [USB3.2r1; 7.5].

    Attributes
    ----------
    link_ready: Signal(), output
        Asserted when the link is ready to perform normal USB operations starting
        at the link layer. When de-asserted, data should not flow above the link layer.
    in_usb_reset: Signal(), input
        Should be asserted whenever a USB reset is detected; including a power-on reset (e.g.
        the PHY has yet to start up), or LFPS warm reset signaling.

    engage_terminations: Signal(), output
        When asserted, the physical layer should be applying link terminations.
    invert_rx_polarity: Signal(). output
        When asserted, indicates that the physical layer should invert all received data at the I/O
        boundary to counter a swapped Rx differential pair.

    perform_rx_detection: Signal(), output
        Asserted to request that the physical layer perform receiver detection, in which the link

    enable_scrambling: Signal(), output
        Asserted when the physical layer should be performing scrambling.


    Parameters
    ----------
    ss_clock_frequency: float
        The frequency of the SS clock domain; used for computing timeouts.
    loosen_requirements: bool
        If True, the requirements will be relaxed from the USB3 specification, in order
        to make things work a little more easily on a variety of PHYs and setups.
    """

    def __init__(self, ss_clock_frequency=125e6, *, loosen_requirements=True):
        self._clock_frequency = ss_clock_frequency
        self._loosen_requirements = loosen_requirements

        #
        # I/O port.
        #
        self.power_on_reset            = Signal()

        self.link_ready                = Signal()
        self.in_usb_reset              = Signal()
        self.entering_u0               = Signal()

        # External event controls.
        self.trigger_link_recovery     = Signal()

        # Power states.
        self.phy_ready                 = Signal()

        # Link control signals.
        self.tx_electrical_idle        = Signal()
        self.engage_terminations       = Signal()
        self.invert_rx_polarity        = Signal()
        self.train_equalizer           = Signal()
        self.disable_scrambling        = Signal()

        # Receiver detection.
        self.perform_rx_detection      = Signal()
        self.link_partner_detected     = Signal()
        self.no_link_partner_detected  = Signal()

        # LFPS detection / emission.
        self.lfps_polling_detected     = Signal()
        self.send_lfps_polling         = Signal()
        self.lfps_cycles_sent          = Signal(16)

        # Training set detection signals.
        self.tseq_detected             = Signal()
        self.ts1_detected              = Signal()
        self.inverted_ts1_detected     = Signal()
        self.ts2_detected              = Signal()

        self.hot_reset_requested       = Signal()
        self.loopback_requested        = Signal()
        self.no_scrambling_requested   = Signal()

        # Training set generation signals.
        self.send_tseq_burst           = Signal()
        self.send_ts1_burst            = Signal()
        self.send_ts2_burst            = Signal()
        self.ts_burst_complete         = Signal()
        self.request_hot_reset         = Signal()
        self.request_no_scrambling     = Signal()

        # Late-stage link management; physical layer control.
        self.enable_scrambling         = Signal()
        self.perform_idle_handshake    = Signal()
        self.idle_handshake_complete   = Signal()

        # Loopback & compliance.
        self.act_as_loopback           = Signal()
        self.emit_compliance_pattern   = Signal()



    def elaborate(self, platform):
        m = Module()

        #
        #  Default signal values.
        #

        # Engage our receive terminations, unless the current state directs otherwise.
        m.d.comb += self.engage_terminations.eq(1)


        #
        # Timeout tracking.
        #

        # Create a timer that can count up to at least 360mS, the largest LTSSM state timeout.
        # [USB 3.2r1: 7.5]
        cycles_in_360mS = int(math.ceil(360e-3 * self._clock_frequency))
        cycles_in_state = Signal(range(cycles_in_360mS + 1))

        # Count by default; this will be automatically cleared on state transitions.
        m.d.ss += cycles_in_state.eq(cycles_in_state + 1)


        #
        # Asynchronous Training Sequence & LFPS Detectors
        #
        polling_seen            = Signal()
        ts2_seen                = Signal()
        hot_reset_seen          = Signal()
        loopback_seen           = Signal()
        disable_scrambling_seen = Signal()
        burst_minimum_met       = Signal()

        with m.If(self.lfps_polling_detected):
            m.d.ss += polling_seen.eq(1)

        with m.If(self.ts2_detected):
            m.d.ss += ts2_seen.eq(1)

        with m.If(self.hot_reset_requested):
            m.d.ss += hot_reset_seen.eq(1)

        with m.If(self.loopback_requested):
            m.d.ss += loopback_seen.eq(1)

        with m.If(self.no_scrambling_requested):
            m.d.ss += disable_scrambling_seen.eq(1)

        with m.If(self.ts_burst_complete):
            m.d.ss += burst_minimum_met.eq(1)


        #
        # FSM helpers.
        #

        # Create a list of tasks to perform on entry to a given state.
        # These will be automatically handled by transition_to_state()
        tasks_on_entry = {}


        def transition_to_state(state):
            """ FSM helper that handles transitions to the given state.

            Automatically handles any "on entry" conditions for the given state.
            """

            # Clear our "time-in-state" counter, and some of our mode flags.
            m.d.ss += [
                cycles_in_state         .eq(0),
                self.request_hot_reset  .eq(0)
            ]

            # If we have any additional entry conditions for the given state, apply them.
            if state in tasks_on_entry:
                m.d.ss += tasks_on_entry[state]

            m.next = state


        def transition_on_timeout(timeout, *, to):
            """ FSM helper that adds a state transition that is automatically invoked after a timeout. """

            # Figure out how many cycles need to pass before we consider ourselves timed out.
            timeout_in_cycles = int(math.ceil(timeout * self._clock_frequency))

            # If we've reached that many cycles, transition to the target state.
            with m.If(cycles_in_state == timeout_in_cycles):
                transition_to_state(to)


        def handle_warm_resets():
            """ FSM helper that automatically moves back to the Rx.Detect.Reset state when appropriate."""

            # If we're in USB reset, we're actively receiving warm reset signaling; and we should reset
            # to the Rx.Detect.Reset state.
            with m.If(self.in_usb_reset):
                transition_to_state("Rx.Detect.Reset")


        #
        # FSM state variables.
        #

        # We'll track if we've seen LFPS bursts during this state;
        # and keep a moving target of how many LFPS bursts we need to see.
        lfps_burst_seen   = Signal()
        target_lfps_count = Signal(16)


        #
        # FSM entry tasks.
        #

        tasks_on_entry['Polling.LFPS'] = [
            lfps_burst_seen    .eq(0),
            target_lfps_count  .eq(16)
        ]

        # Ensure we enter Polling.Active with fresh device state.
        tasks_on_entry['Polling.Active'] = [
            ts2_seen                  .eq(0),
            hot_reset_seen            .eq(0),
            loopback_seen             .eq(0),
            disable_scrambling_seen   .eq(0),
            self.request_no_scrambling.eq(self.disable_scrambling),
            burst_minimum_met         .eq(0)
        ]


        # Clear our previous training state on entering recovery.
        tasks_on_entry['Recovery.Active'] = [
            ts2_seen                  .eq(0),
            hot_reset_seen            .eq(0),
            loopback_seen             .eq(0),
            disable_scrambling_seen   .eq(0),
            self.request_no_scrambling.eq(self.disable_scrambling),
            burst_minimum_met         .eq(0)
        ]

        # Ensure we enter our primary training sequence with a fresh view of whether we've
        # seen any of our training sets.
        tasks_on_entry['Polling.RxEQ'] = [
            ts2_seen                  .eq(0),
            hot_reset_seen            .eq(0),
            disable_scrambling_seen   .eq(0),
            self.request_no_scrambling.eq(self.disable_scrambling),
        ]


        # Clear our previous training state on entering recovery.
        tasks_on_entry['Hot Reset.Active'] = [
            ts2_seen                  .eq(0),
            self.request_hot_reset    .eq(1)
        ]



        #
        # Main Link Training and Status State Machine
        #
        with m.FSM(domain="ss"):

            # Rx.Detect.Reset -- we've just started link bringup post-reset; and are ready to
            # perform any necessary link configuration.
            with m.State("Rx.Detect.Reset"):
                m.d.comb += [

                    # Keep ourselves from transmitting until we're ready to send...
                    self.tx_electrical_idle   .eq(1),

                    # ... and prevent ourselves from presenting receiver terminations until
                    # our PHY has started up; so the other side doesn't start LFPS polling, yet.
                    self.engage_terminations  .eq(0)
                ]


                # We'll wait in this state until our PHY is brought up, and we're not detecting
                # any Warm Reset LFPS signaling.
                with m.If(~self.in_usb_reset & self.phy_ready):
                    transition_to_state("Rx.Detect.Active")


            # Rx.Detect.Active -- we're now post-reset; and we're going to attempt to detect a
            # link partner, by checking for far-end receiver terminations. Basically, we try to
            # detect whether we're connected to another SuperSpeed transciever via a cable, so
            # we don't waste time performing link training if our link isn't there.
            with m.State("Rx.Detect.Active"):
                m.d.comb += [
                    self.tx_electrical_idle    .eq(1),
                    self.perform_rx_detection  .eq(1)
                ]

                with m.If(self.link_partner_detected):
                    transition_to_state("Polling.LFPS")
                with m.If(self.no_link_partner_detected):
                    transition_to_state("Rx.Detect.Quiet")


            # Rx.Detect.Quiet -- we've performed a link detection, but didn't detect anyone.
            # We'll wait here until our next detection cycle, saving the power of performing
            # continuous detections.
            with m.State("Rx.Detect.Quiet"):
                m.d.comb += self.tx_electrical_idle.eq(1)

                # TODO: count our number of failed attempts; and disable
                # SuperSpeed after we see eight of them.

                # After 12ms, try again.
                transition_on_timeout(12e-3, to="Rx.Detect.Active")


            # Polling.LFPS -- now that we know there's someone listening on the other side, we'll
            # begin exchanging LFPS messages; giving the two sides the opportunity to sync up and
            # establish initial DC characteristics. [USB 3.2r1: 7.5.4.3]
            with m.State("Polling.LFPS"):
                m.d.comb += self.tx_electrical_idle.eq(1)


                # Continuously send our LFPS polling.
                m.d.comb += self.send_lfps_polling.eq(1)

                # To move forward with the LTSSM, we'll need to:
                # - Have sent at least 16 bursts.
                # - Have transmitted at least four bursts since we first saw a burst.
                with m.If(self.lfps_cycles_sent >= target_lfps_count):

                    # If we see a TS1, and we're not in strict mode, move forward without
                    # necessarily seeing a LFPS burst ourselves.
                    with m.If(self._loosen_requirements & self.ts1_detected):
                            transition_to_state("Polling.RxEQ")

                    # If this is the first burst we've seen, move our target forward;
                    # so we can meet our second condition.
                    with m.If(self.lfps_polling_detected & ~lfps_burst_seen):
                        m.d.ss += [
                            lfps_burst_seen    .eq(1),
                            target_lfps_count  .eq(self.lfps_cycles_sent + 4)
                        ]

                    # If we've sent enough, -and- we meet our condition, move forward.
                    with m.If(lfps_burst_seen):
                            transition_to_state("Polling.RxEQ")


                # If we haven't yet sent 16 bursts, track how many bursts we have sent.
                with m.Elif(self.lfps_polling_detected & ~lfps_burst_seen):
                    m.d.ss += lfps_burst_seen.eq(1)

                    with m.If(self.lfps_cycles_sent > 12):
                        m.d.ss += target_lfps_count.eq(self.lfps_cycles_sent + 4)



                # If we've never seen polling, we'll exit to Compliance once this passes. [USB 3.2r1: 7.5.4.3]
                with m.If(~polling_seen):
                    transition_on_timeout(360e-3, to="Compliance")
                with m.Else():
                    transition_on_timeout(360e-3, to="SS.Disabled.Default")



            # Polling.RxEQ -- we've now seen the other side of our link, and are ready to initialize
            # communications. We'll bring our link online, start sending our first Training Set (TSEQ),
            # and give the PHY time to achieve DC equalization.
            # [USB 3.2.r1: 7.5.4.7]
            with m.State("Polling.RxEQ"):
                handle_warm_resets()

                # Continuously send TSEQs; these are used to perform receiver equalization training.
                m.d.comb += self.send_tseq_burst.eq(1)

                # Request our physical layer to perform equalization training.
                m.d.comb += self.train_equalizer.eq(1)

                # Once we've sent a full burst of 65536 TSEQs, we can begin our link training handshake.
                with m.If(self.ts_burst_complete):
                    transition_to_state("Polling.Active")


            # Polling.Active -- we've now exchanged our initial training sequences, and we're ready to
            # begin exchaning our core training sequences. We'll start sending TS1, and let the PHY handle
            # link training until it reliably the same thing from the host. [USB 3.2r1: 7.5.4.8]
            with m.State("Polling.Active"):
                handle_warm_resets()

                # Constantly send TS1s; which indicate that we're in link training, but haven't yet
                # seen enough TS1s to move forward with training.
                m.d.comb += self.send_ts1_burst.eq(1)

                # If we don't achieve link training within 12mS, we'll assume that we've lost our
                # link partner. We'll start our process again from the beginning.
                transition_on_timeout(12e-3, to="Rx.Detect.Active")

                #
                # The specification allows us to move on to Polling.Configuration as soon as we
                # see a sufficient burst of TS1s, as theoretically a link partner should be able to
                # able to accept TS2s without seeing TS1s [USB3.2r1: 7.5.4.8]; however, experientially
                # many link partners get upset if they don't see at least -some- TS1s.
                #
                # Sending at least 16 total TS1s seems to work around this problem, while still allowing
                # us to move on to sending TS2s in a timely manner.
                #
                with m.If(burst_minimum_met):

                    # Once our ordered set module reports a long enough burst of seen training sets
                    # we're satisfied with our link training. However, we don't want to stop sending
                    # training data until we're sure our link partner has completed training; so we'll
                    # move to Polling.Configuration to await completion of the other side.
                    with m.If(self.ts1_detected | self.ts2_detected):
                        m.d.ss += self.invert_rx_polarity.eq(0),
                        transition_to_state("Polling.Configuration")


                    # If we see a long enough burst of -inverted- training sets, we're also satisfied
                    # with our link training; but we know that the receive differential pair is inverted.
                    # We'll continue, but ask our physical layer to invert our received data.
                    with m.If(self.inverted_ts1_detected):
                        m.d.ss += self.invert_rx_polarity.eq(1),
                        transition_to_state("Polling.Configuration")


            # Polling.Configuration -- we're now satisfied with our link training; we'll need to communicate
            # this to the other side, and wait for the other side to advertise the same. [USB3.2r1; 7.5.4.9]
            with m.State("Polling.Configuration"):
                handle_warm_resets()

                # Constantly send TS2s, which both allow the other side to continue link training and
                # advertise that our side has completed link training itself.
                m.d.comb += self.send_ts2_burst.eq(1)

                # If we don't achieve link training within 12mS, we'll assume that we've lost our
                # link partner. We'll start our process again from the beginning.
                transition_on_timeout(12e-3, to="Rx.Detect.Active")

                # If we've finished sending the requisite amount of TS2s and we've seen TS2s from the
                # other side, we know that both sides are finished with the core link training.
                # Move on to our final
                with m.If(self.ts_burst_complete & ts2_seen):
                    transition_to_state("Polling.Configuration.Exit")


            # Polling.Configuration.Exit [synthetic state; not from the specification] -- once we're
            # satisfied with our TS1/TS2 exchange, we're required to send at least 16 more TS2s, to ensure
            # that the other side sees enough TS2s to know that we're both done. In this state, we'll send
            # a burst of TS2s.
            with m.State("Polling.Configuration.Exit"):
                handle_warm_resets()

                # Continue to send TS2s...
                m.d.comb += self.send_ts2_burst.eq(1)

                # ... until we've sent a full burst of 16; at which point we can advance.
                with m.If(self.ts_burst_complete):
                    transition_to_state("Polling.Idle")


            # Polling.Idle -- we've now finished link training, and we're ready to move on to real
            # communications. We'll perform one final sanity check, and then move to our next state.
            # [USB3.2r1: 7.5.4.10]
            with m.State("Polling.Idle"):
                handle_warm_resets()

                m.d.comb += [
                    # From this state onward, we have an active link, and we can thus enable data scrambling.
                    self.enable_scrambling       .eq(~self.request_no_scrambling & ~disable_scrambling_seen),

                    # Generate our IDL handshake.
                    self.perform_idle_handshake  .eq(1)
                ]

                # If a hot-reset is being requested, we'll enter Hot Reset.Active.
                with m.If(hot_reset_seen):
                    transition_to_state("Hot Reset.Active")

                # If Loopback is being requested, we'll enter Loopback mode.
                with m.Elif(loopback_seen):
                    transition_to_state("Loopback")

                # Otherwise, As one final synchronization step and sanity check, we'll require a proper
                # period of # Logical Idle to be detected before we move to our next state. Since Logical
                # Idle signals are scrambled, this helps to ensure that both sides of the link have
                # synchronized scrambler state and that the other side has stopped sending TS2s.
                with m.Elif(self.idle_handshake_complete):
                    m.d.comb += self.entering_u0.eq(1)
                    transition_to_state("U0")

                # If we don't see that logical idle within 2ms, something's gone wrong. We'll need to
                # start our connection process from the beginning.
                transition_on_timeout(2e-3, to="Rx.Detect.Reset")


            # U0 -- our primary active USB state, in which we've completed link bringup and now are
            # performing normal USB3 operations.
            with m.State("U0"):
                handle_warm_resets()

                m.d.comb += [
                    # We're now ready for normal operation -- we'll mark our link as ready,
                    # and keep our normal scrambling enabled.
                    self.enable_scrambling  .eq(~self.request_no_scrambling & ~disable_scrambling_seen),
                    self.link_ready         .eq(1)
                ]

                # If we've seen an event that requires link recovery, move into link recovery.
                with m.If(self.trigger_link_recovery):
                    transition_to_state("Recovery.Active")

                # If we've seen a TS1 ordered set, we know the other side has gone into recovery.
                # We should, as well.
                with m.If(self.ts1_detected):
                    transition_to_state("Recovery.Active")


                # TODO: handle the various other cases for leaving U0

            # Hot Reset.Active -- during link training, we've seen a training set indicating
            # we should perform a hot reset. We're now performing a TS2 handshake, modified so
            # we are also sending Hot Reset.
            with m.State("Hot Reset.Active"):
                handle_warm_resets()

                # As in Polling.Configuration, we'll send TS2s; but we'll send them with our
                # Hot Reset bit set.
                m.d.comb += [
                    self.send_ts2_burst     .eq(1),
                ]

                # If we don't achieve link training within 12mS, we'll assume that we've lost our
                # link partner. We'll assume our link is no longer recoverable, and move to inactive.
                transition_on_timeout(12e-3, to="SS.Inactive.Quiet")

                # We need to at least one burst with the Reset bit set; and then drop out of hot reset.
                with m.If(self.ts_burst_complete):
                    m.d.ss += self.request_hot_reset.eq(0)

                # Once we've seen TS2s in response that don't have Hot Reset asserted, we can drop out
                # of hot reset; and pursue normal operation again.
                with m.If(self.ts_burst_complete & ts2_seen & ~self.hot_reset_requested):
                    transition_to_state("Hot Reset.Exit")


            # Hot Reset.Exit -- we've now finished link training, and we're ready to move on to having
            # an active link. We'll now perform a reduced-complexity Idle handshake.
            with m.State("Hot Reset.Exit"):
                handle_warm_resets()

                m.d.comb += [
                    # From this state onward, we have an active link, and we can thus enable data scrambling.
                    self.enable_scrambling       .eq(~self.request_no_scrambling & ~disable_scrambling_seen),

                    # Generate our IDL handshake.
                    self.perform_idle_handshake  .eq(1)
                ]

                # Once we've finished our Idle handshake, we can move on to U0.
                with m.If(self.idle_handshake_complete):
                    m.d.comb += self.entering_u0.eq(1)
                    transition_to_state("U0")

                # If we don't complete our Idle handshake within 2ms, something's gone wrong.
                # We'll consider our link irrecoverable.
                transition_on_timeout(2e-3, to="SS.Inactive.Quiet")


            # Recovery.Active -- our link is no longer in a reliably usable state; we'll need
            # to perform a re-training before we can use it fully. However, since we've already
            # performed our initial receiver equalization, we can maintain its settings and perform
            # only the last steps of training.
            with m.State("Recovery.Active"):
                handle_warm_resets()

                # As in Polling.Active, we'll send TS1s to establish training.
                m.d.comb += self.send_ts1_burst.eq(1)

                # If we don't achieve link training within 12mS, we'll assume that we've lost our
                # link partner. We'll assume our link is no longer recoverable, and move to inactive.
                transition_on_timeout(12e-3, to="SS.Inactive.Quiet")

                #
                # The specification allows us to move on to Polling.Configuration as soon as we
                # see a sufficient burst of TS1s, as theoretically a link partner should be able to
                # able to accept TS2s without seeing TS1s [USB3.2r1: 7.5.10.3]; however, experientially
                # many link partners get upset if they don't see at least -some- TS1s.
                #
                # Sending at least 16 total TS1s seems to work around this problem, while still allowing
                # us to move on to sending TS2s in a timely manner.
                #
                with m.If(burst_minimum_met):

                    # Once we see enough TS1s from the other side; or see TS2s, we'll move into our next step.
                    with m.If(self.ts1_detected | self.ts2_detected):
                        transition_to_state("Recovery.Configuration")


            # Recovery.Configuration -- we're now satisfied with our link training; we'll need to communicate
            # this to the other side, and wait for the other side to advertise the same. [USB3.2r1; 7.5.4.9]
            with m.State("Recovery.Configuration"):
                handle_warm_resets()

                # Constantly send TS2s.
                m.d.comb += self.send_ts2_burst.eq(1)

                # If we don't achieve link training within 12mS, we'll assume that we've lost our
                # link partner. We'll assume our link is no longer up, and move to inactive.
                transition_on_timeout(12e-3, to="SS.Inactive.Quiet")

                # If we've finished sending the requisite amount of TS2s and we've seen TS2s from the
                # other side, we know that both sides are finished with the core link training.
                # Move on to our final
                with m.If(self.ts_burst_complete & ts2_seen):
                    transition_to_state("Recovery.Configuration.Exit")


            # Recovery.Configuration.Exit [synthetic state; not from the specification] -- once we're
            # satisfied with our TS1/TS2 exchange, we're required to send at least 16 more TS2s, to ensure
            # that the other side sees enough TS2s to know that we're both done. In this state, we'll send
            # a burst of TS2s.
            with m.State("Recovery.Configuration.Exit"):
                handle_warm_resets()

                # Continue to send TS2s...
                m.d.comb += self.send_ts2_burst.eq(1)

                # ... until we've sent a full burst of 16; at which point we can advance.
                with m.If(self.ts_burst_complete):
                    transition_to_state("Recovery.Idle")


            # Recovery.Idle -- we've now finished link re-training; and are waiting to see that the other
            # side has also finished sending TS2s [USB3.2r1: 7.5.4.10].
            with m.State("Recovery.Idle"):
                handle_warm_resets()

                m.d.comb += [
                    # Restore scrambling, and repeat our idle handshake.
                    self.enable_scrambling       .eq(~self.request_no_scrambling & ~disable_scrambling_seen),
                    self.perform_idle_handshake  .eq(1)
                ]

                # If a hot-reset is being requested, we'll enter Hot Reset.Active.
                with m.If(hot_reset_seen):
                    transition_to_state("Hot Reset.Active")

                # If Loopback is being requested, we'll enter Loopback mode.
                with m.Elif(loopback_seen):
                    transition_to_state("Loopback")

                # Otherwise, As one final synchronization step and sanity check, we'll require a proper
                # period of Logical Idle to be detected before we move to our next state. Since Logical
                # Idle signals are scrambled, this helps to ensure that both sides of the link have
                # synchronized scrambler state and that the other side has stopped sending TS2s.
                with m.Elif(self.idle_handshake_complete):
                    m.d.comb += self.entering_u0.eq(1)
                    transition_to_state("U0")

                # If we don't see that logical idle within 2ms, something's gone wrong. We'll
                # assume we've lost our link partner, and move to SS.Inactive.
                transition_on_timeout(2e-3, to="SS.Inactive.Quiet")


            # Compliance -- we've failed link training in such a way as to believe we're in the
            # middle of a compliance test / validation (lucky us!).
            with m.State("Compliance"):
                handle_warm_resets()

                # We don't currently handle Compliance properly. In this case, this message refers
                # to the Compliance state, but this also makes us non-compliant, so this statement
                # has an especially appropriate double meaning.
                #
                # We'll throw our hands up in despair and re-try link training.
                # Maybe this time it'll work.
                transition_to_state("Rx.Detect.Reset")


            # Loopback -- during the link bringup, our link partner requested that we go into
            # Loopback mode; so we'll begin acting as a loopback device.
            with m.State("Loopback"):
                handle_warm_resets()
                m.d.comb += self.act_as_loopback.eq(1)

                # FIXME: detect Loopback Exit LFPS, and exit this state.


            # SS.Inactive.Quiet -- an non-recoverable error has occurred somewhere with the link.
            # We'll wait for a bit here and do nothing, so we don't completely drain power.
            with m.State("SS.Inactive.Quiet"):
                handle_warm_resets()

                m.d.comb += self.tx_electrical_idle.eq(1),
                transition_on_timeout(12e-3, to="SS.Inactive.Disconnect.Detect")


            # SS.Inactive.Disconnect.Detect  -- our best case scenario is that we become disconnected,
            # and then are reconnected to establish a working link. We'll check for disconnection.
            with m.State("SS.Inactive.Disconnect.Detect"):
                handle_warm_resets()

                m.d.comb += [
                    self.tx_electrical_idle    .eq(1),
                    self.perform_rx_detection  .eq(1)
                ]

                # If we detect a link partner, we're still in our non-recoverable state.
                # We'll go back to .Quiet and wait another 12ms to check again.
                with m.If(self.link_partner_detected):
                    transition_to_state("SS.Inactive.Quiet")

                # If we detect the absence of a link partner, we're no longer in our bad state.
                # We'll move to Rx.Detect, and start over again.
                with m.If(self.no_link_partner_detected):
                    transition_to_state("Rx.Detect.Quiet")


            # SS.Disabled.Default -- the SuperSpeed portion of our link is disabled; we'll remove
            # our terminations and attempt to act as a valid USB2 device.
            with m.State("SS.Disabled.Default"):
                handle_warm_resets()

                m.d.comb += [
                    self.tx_electrical_idle    .eq(1),
                    self.engage_terminations   .eq(0)
                ]

                # FIXME: transition to SS.Disabled.Error if we get here three times without success.


            # SS.Disabled.Error -- the SuperSpeed portion of our link is disabled; we'll remove
            # our terminations and sit idly until VBUS is cycled.
            with m.State("SS.Disabled.Error"):
                handle_warm_resets()

                m.d.comb += [
                    self.tx_electrical_idle    .eq(1),
                    self.engage_terminations   .eq(0)
                ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test               import LunaSSGatewareTestCase, ss_domain_test_case

from luna.gateware.usb.usb3.application.request import SuperSpeedSetupDecoder
from usb_protocol.types     import USBRequestType, USBRequestRecipient

class SuperSpeedSetupDecoderTest(LunaSSGatewareTestCase):
    FRAGMENT_UNDER_TEST = SuperSpeedSetupDecoder

    @ss_domain_test_case
    def test_setup_parse(self):
        dut   = self.dut
        sink  = dut.sink
        setup = dut.packet

        # Mark our data as always valid SETUP data.
        yield dut.header_in.setup.eq(1)
        yield sink.valid.eq(0b1111)

        # Provide our first word...
        yield sink.first.eq(1)
        yield sink.last .eq(0)
        yield sink.data .eq(0x2211AAC1)
        yield

        # ... then our second ...
        yield sink.first.eq(0)
        yield sink.last .eq(1)
        yield sink.data .eq(0x00043344)
        yield

        # ... then mark our packet as good.
        yield from self.pulse(dut.rx_good)
        yield

        # Finally, check that our fields have been parsed properly.
        self.assertEqual((yield setup.is_in_request), 1)
        self.assertEqual((yield setup.type),          USBRequestType.VENDOR)
        self.assertEqual((yield setup.recipient),     USBRequestRecipient.INTERFACE)
        self.assertEqual((yield setup.request),       0xAA)
        self.assertEqual((yield setup.value),         0x2211)
        self.assertEqual((yield setup.index),         0x3344)
        self.assertEqual((yield setup.length),        4)




#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Endpoint interfaces for providing status updates to the host.

These are mainly meant for use with interrupt endpoints; and allow a host to e.g.
repeatedly poll a device for status.
"""

from amaranth         import Elaboratable, Module, Signal, Array
from amaranth.lib.cdc import FFSynchronizer

from ..endpoint     import EndpointInterface


class USBSignalInEndpoint(Elaboratable):
    """ Endpoint that transmits the value of a signal to a host whenever polled.

    This is intended to be usable to implement a simple interrupt endpoint that polls for a status signal.

    Attributes
    ----------
    signal: Signal(<variable width>), input
        The signal to be relayed to the host. This signal's current value will be relayed each time the
        host polls our endpoint.
    interface: EndpointInterface
        Communications link to our USB device.

    status_read_complete: Signal(), output
        Strobe that pulses high for a single `usb`-domain cycle each time a status read is complete.

    Parameters
    ----------
    width: int
        The width of the signal we'll relay up to the host, in bits.
    endpoint_number: int
        The endpoint number (not address) this endpoint should respond to.
    endianness: str, "big" or "little", optional
        The endianness with which to send the data. Defaults to little endian.
    signal_domain: str, optional
        The name of the domain :attr:``signal`` is clocked from. If this value is anything other than
        "usb", the signal will automatically be synchronized to the USB clock domain.
    """

    def __init__(self, *, width, endpoint_number, endianness="little", signal_domain="usb"):
        self._width           = width
        self._endpoint_number = endpoint_number
        self._signal_domain   = signal_domain
        self._endianness      = endianness

        if self._endianness not in ("big", "little"):
            raise ValueError(f"Endianness must be 'big' or 'little', not {endianness}.")

        #
        # I/O port
        #
        self.signal               = Signal(self._width)
        self.interface            = EndpointInterface()

        self.status_read_complete = Signal()



    def elaborate(self, platform):
        m = Module()

        # Shortcuts.
        tx = self.interface.tx
        tokenizer = self.interface.tokenizer


        # Grab a copy of the relevant signal that's in our USB domain; synchronizing if we need to.
        if self._signal_domain == "usb":
            target_signal = self.signal
        else:
            target_signal = Signal()
            m.submodules.target_signal_cdc = FFSynchronizer(self.signal, target_signal, o_domain="usb")


        # Store a latched version of our signal, captured before we start a transmission.
        latched_signal = Signal.like(self.signal)

        # Grab an byte-indexable reference into our signal.
        bytes_in_signal = (self._width + 7) // 8
        signal_bytes = Array(latched_signal[n * 8 : n * 8 + 8] for n in range(bytes_in_signal))

        # Store how many bytes we've transmitted.
        bytes_transmitted = Signal(range(0, bytes_in_signal + 1))

        #
        # Data transmission logic.
        #

        # If this signal is big endian, send them in reading order; otherwise, index our multiplexer in reverse.
        # Note that our signal is captured little endian by default, due the way we use Array() above. If we want
        # big endian; then we'll flip it.
        if self._endianness == "little":
            index_to_transmit = bytes_transmitted
        else:
            index_to_transmit = bytes_in_signal - bytes_transmitted - 1

        # Always transmit the part of the latched signal byte that corresponds to our
        m.d.comb += tx.payload.eq(signal_bytes[index_to_transmit])

        #
        # Core control FSM.
        #

        endpoint_number_matches  = (tokenizer.endpoint == self._endpoint_number)
        targeting_endpoint       = endpoint_number_matches & tokenizer.is_in
        packet_requested         = targeting_endpoint & tokenizer.ready_for_response


        with m.FSM(domain="usb"):

            # IDLE -- we've not yet gotten an token requesting data. Wait for one.
            with m.State('IDLE'):

                # Once we're ready to send a response...
                with m.If(packet_requested):

                    m.d.usb += [
                        # ... clear our transmit counter ...
                        bytes_transmitted  .eq(0),

                        # ... latch in our response...
                        latched_signal     .eq(self.signal),
                    ]

                    # ...  and start transmitting it.
                    m.next = "TRANSMIT_RESPONSE"


            # TRANSMIT_RESPONSE -- we're now ready to send our latched response to the host.
            with m.State("TRANSMIT_RESPONSE"):
                is_last_byte = bytes_transmitted + 1 == bytes_in_signal

                # While we're transmitting, our Tx data is valid.
                m.d.comb += [
                    tx.valid  .eq(1),
                    tx.first  .eq(bytes_transmitted == 0),
                    tx.last   .eq(is_last_byte)
                ]

                # Each time we receive a byte, move on to the next one.
                with m.If(tx.ready):
                    m.d.usb += bytes_transmitted.eq(bytes_transmitted + 1)

                    # If this is the last byte to be transmitted, move to waiting for an ACK.
                    with m.If(is_last_byte):
                        m.next = "WAIT_FOR_ACK"


            # WAIT_FOR_ACK -- we've now transmitted our full packet; we need to wait for the host to ACK it
            with m.State("WAIT_FOR_ACK"):

                # If the host does ACK, we're done! Move back to our idle state.
                with m.If(self.interface.handshakes_in.ack):
                    m.d.comb += self.status_read_complete.eq(1)
                    m.d.usb += self.interface.tx_pid_toggle[0].eq(~self.interface.tx_pid_toggle[0])
                    m.next = "IDLE"


                # If the host starts a new packet without ACK'ing, we'll need to retransmit.
                # Wait for a new IN token.
                with m.If(self.interface.tokenizer.new_token):
                    m.next = "RETRANSMIT"


            # RETRANSMIT -- the host failed to ACK the data we've most recently sent.
            # Wait here for the host to request the data again.
            with m.State("RETRANSMIT"):

                # Once the host does request the data again...
                with m.If(packet_requested):

                    # ... retransmit it, starting from the beginning.
                    m.d.usb += bytes_transmitted.eq(0),
                    m.next = "TRANSMIT_RESPONSE"

        return m





#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test           import LunaGatewareTestCase, usb_domain_test_case

from luna.gateware.usb.usb2       import USBSpeed
from luna.gateware.usb.usb2.reset import USBResetSequencer
from luna.gateware.interface.utmi     import UTMIOperatingMode, UTMITerminationSelect

class USBResetSequencerTest(LunaGatewareTestCase):
    FRAGMENT_UNDER_TEST = USBResetSequencer

    SYNC_CLOCK_FREQUENCY = None
    USB_CLOCK_FREQUENCY  = 60e6

    def instantiate_dut(self):
        dut = super().instantiate_dut()

        # Test tweak: squish down our delays to speed up sim.
        dut._CYCLES_2P5_MICROSECONDS = 10

        return dut


    def initialize_signals(self):

        # Start with a non-reset line-state.
        yield self.dut.line_state.eq(0b01)

    @usb_domain_test_case
    def test_full_speed_reset(self):
        dut = self.dut

        yield from self.advance_cycles(10)

        # Before we detect a reset, we should be at normal FS,
        # and we should be in reset until VBUS is provided.
        self.assertEqual((yield dut.bus_reset),          1)
        self.assertEqual((yield dut.current_speed),      USBSpeed.FULL)
        self.assertEqual((yield dut.operating_mode),     UTMIOperatingMode.NORMAL)
        self.assertEqual((yield dut.termination_select), UTMITerminationSelect.LS_FS_NORMAL)

        # Once we apply VBUS, we should drop out of reset...
        yield dut.vbus_connected.eq(1)
        yield
        self.assertEqual((yield dut.bus_reset), 0)

        # ... and stay that way.
        yield from self.advance_cycles(dut._CYCLES_2P5_MICROSECONDS)
        self.assertEqual((yield dut.bus_reset), 0)

        yield dut.line_state.eq(0)

        # After assertion of SE0, we should remain out of reset for >2.5us, <3ms...
        yield
        self.assertEqual((yield dut.bus_reset), 0)

        # ... we choose to wait for 5us and then we should see a cycle of reset.
        yield from self.advance_cycles(dut._CYCLES_5_MICROSECONDS)
        self.assertEqual((yield dut.bus_reset), 1)

        yield from self.advance_cycles(10)
        yield dut.line_state.eq(0b01)
        yield

        # Finally, we should arrive in HS_DETECT, post-reset.
        self.assertEqual((yield dut.current_speed),      USBSpeed.HIGH)
        self.assertEqual((yield dut.operating_mode),     UTMIOperatingMode.CHIRP)
        self.assertEqual((yield dut.termination_select), UTMITerminationSelect.HS_CHIRP)


    #
    # It would be lovely to have tests that run through each of our reset/suspend
    # cases here; but currently the time it takes run through the relevant delays is
    # prohibitive. :(
    #



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Clock and reset (CAR) controllers for LUNA. """

import logging

from abc import ABCMeta, abstractmethod
from amaranth import Signal, Module, ClockDomain, ClockSignal, Elaboratable, Instance, ResetSignal

from ..utils.cdc import stretch_strobe_signal


class PHYResetController(Elaboratable):
    """ Gateware that implements a short power-on-reset pulse to reset an attached PHY.

    I/O ports:

        I: trigger   -- A signal that triggers a reset when high.
        O: phy_reset -- The signal to be delivered to the target PHY.
    """

    def __init__(self, *, clock_frequency=60e6, reset_length=2e-6, stop_length=2e-6, power_on_reset=True):
        """ Params:

            reset_length   -- The length of a reset pulse, in seconds.
            stop_length    -- The length of time STP should be asserted after reset.
            power_on_reset -- If True or omitted, the reset will be applied once the firmware
                              is configured.
        """

        from math import ceil

        self.power_on_reset = power_on_reset

        # Compute the reset length in cycles.
        clock_period = 1 / clock_frequency
        self.reset_length_cycles = ceil(reset_length / clock_period)
        self.stop_length_cycles  = ceil(stop_length  / clock_period)

        #
        # I/O port
        #
        self.trigger   = Signal()
        self.phy_reset = Signal()
        self.phy_stop  = Signal()


    def elaborate(self, platform):
        m = Module()

        # Counter that stores how many cycles we've spent in reset.
        cycles_in_reset = Signal(range(0, self.reset_length_cycles))

        reset_state = 'RESETTING' if self.power_on_reset else 'IDLE'
        with m.FSM(init=reset_state, domain='sync') as fsm:

            # Drive the PHY reset whenever we're in the RESETTING cycle.
            m.d.comb += [
                self.phy_reset.eq(fsm.ongoing('RESETTING')),
                self.phy_stop.eq(~fsm.ongoing('IDLE'))
            ]

            with m.State('IDLE'):
                m.d.sync += cycles_in_reset.eq(0)

                # Wait for a reset request.
                with m.If(self.trigger):
                    m.next = 'RESETTING'

            # RESETTING: hold the reset line active for the given amount of time
            with m.State('RESETTING'):
                m.d.sync += cycles_in_reset.eq(cycles_in_reset + 1)

                with m.If(cycles_in_reset + 1 == self.reset_length_cycles):
                    m.d.sync += cycles_in_reset.eq(0)
                    m.next = 'DEFERRING_STARTUP'

            # DEFERRING_STARTUP: Produce a signal that will defer startup for
            # the provided amount of time. This allows line state to stabilize
            # before the PHY will start interacting with us.
            with m.State('DEFERRING_STARTUP'):
                m.d.sync += cycles_in_reset.eq(cycles_in_reset + 1)

                with m.If(cycles_in_reset + 1 == self.stop_length_cycles):
                    m.d.sync += cycles_in_reset.eq(0)
                    m.next = 'IDLE'


        return m



class LunaDomainGenerator(Elaboratable, metaclass=ABCMeta):
    """ Helper that generates the clock domains used in a LUNA board.

    Note that this module should create three in-phase clocks; so these domains
    should not require explicit boundary crossings.

    I/O port:
        O: clk_fast      -- The clock signal for our fast clock domain.
        O: clk_sync      -- The clock signal used for our sync clock domain.
        O: clk_usb       -- The clock signal used for our USB domain.
        O: usb_holdoff   -- Signal that indicates that the USB domain is immediately post-reset,
                            and thus we should avoid transactions with the external PHY.
    """

    def __init__(self, *, clock_signal_name=None, clock_signal_frequency=None):
        """
        Parameters:
            clock_signal_name      = The clock signal name to use; or None to use the platform's default clock.
            clock_signal_frequency = The frequency of clock_signal_name in Hz.
        """

        self.clock_name      = clock_signal_name
        self.clock_frequency = clock_signal_frequency

        #
        # I/O port
        #
        self.clk_fast     = Signal()
        self.clk_sync     = Signal()
        self.clk_usb      = Signal()

        self.usb_holdoff  = Signal()


    @abstractmethod
    def generate_fast_clock(self, m, platform):
        """ Method that returns our platform's fast clock; used for e.g. RAM interfacing. """


    @abstractmethod
    def generate_sync_clock(self, m, platform):
        """ Method that returns our platform's primary synchronous clock. """


    @abstractmethod
    def generate_usb_clock(self, m, platform):
        """ Method that generates a 60MHz clock used for ULPI interfacing. """


    def create_submodules(self, m, platform):
        """ Method hook for creating any necessary submodules before generating clock. """
        pass


    def create_usb_reset(self, m, platform):
        """
        Function that should create our USB reset, and connect it to:
            m.domains.usb.rst / self.usb_rst
        """

        m.submodules.usb_reset = controller = PHYResetController()
        m.d.comb += [
            ResetSignal("usb")  .eq(controller.phy_reset),
            self.usb_holdoff    .eq(controller.phy_stop)
        ]


    def elaborate(self, platform):
        m = Module()

        # Create our clock domains.
        m.domains.fast = self.fast = ClockDomain()
        m.domains.sync = self.sync = ClockDomain()
        m.domains.usb  = self.usb  = ClockDomain()

        # Call the hook that will create any submodules necessary for all clocks.
        self.create_submodules(m, platform)

        # Generate and connect up our clocks.
        m.d.comb += [
            self.clk_usb                   .eq(self.generate_usb_clock(m, platform)),
            self.clk_sync                  .eq(self.generate_sync_clock(m, platform)),
            self.clk_fast                  .eq(self.generate_fast_clock(m, platform)),

            ClockSignal(domain="fast")     .eq(self.clk_fast),
            ClockSignal(domain="sync")     .eq(self.clk_sync),
            ClockSignal(domain="usb")      .eq(self.clk_usb),
        ]

        # Call the hook that will connect up our reset signals.
        self.create_usb_reset(m, platform)

        return m


class LunaECP5DomainGenerator(LunaDomainGenerator):
    """ ECP5 clock domain generator for LUNA. Assumes a 60MHz input clock. """

    # For debugging, we'll allow the ECP5's onboard clock to generate a 62MHz
    # oscillator signal. This won't work for USB, but it'll at least allow
    # running some basic self-tests. The clock is 310 MHz by default, so
    # dividing by 5 will yield 62MHz.
    OSCG_DIV = 5

    # Quick configuration selection
    DEFAULT_CLOCK_FREQUENCIES_MHZ = {
        "fast": 240,
        "sync": 120,
        "usb":  60
    }

    def __init__(self, *, clock_frequencies=None, clock_signal_name=None, clock_signal_frequency=None):
        """
        Parameters:
            clock_frequencies -- A dictionary mapping 'fast', 'sync', and 'usb' to the clock
                                 frequencies for those domains, in MHz. Valid choices for each
                                 domain are 60, 120, and 240. If not provided, fast will be
                                 assumed to be 240, sync will assumed to be 120, and usb will
                                 be assumed to be a standard 60.
        """
        super().__init__(clock_signal_name=clock_signal_name, clock_signal_frequency=clock_signal_frequency)
        self.clock_frequencies = clock_frequencies


    def create_submodules(self, m, platform):

        self._pll_lock   = Signal()


        # Figure out our platform's clock frequencies -- grab the platform's
        # defaults, and then override any with our local, caller-provided copies.
        new_clock_frequencies = platform.DEFAULT_CLOCK_FREQUENCIES_MHZ.copy()
        if self.clock_frequencies:
            new_clock_frequencies.update(self.clock_frequencies)
        self.clock_frequencies = new_clock_frequencies


        # Use the provided clock name and frequency for our input; or the default clock
        # if no name was provided.
        clock_name = self.clock_name if self.clock_name else platform.default_clk
        clock_frequency = self.clock_frequency if self.clock_name else platform.default_clk_frequency

        # Create absolute-frequency copies of our PLL outputs.
        # We'll use the generate_ methods below to select which domains
        # apply to which components.
        self._clk_240MHz = Signal()
        self._clk_120MHz = Signal()
        self._clk_60MHz  = Signal()
        self._clock_options = {
            60:  self._clk_60MHz,
            120: self._clk_120MHz,
            240: self._clk_240MHz
        }

        pll_params = {}

        # Grab our input clock
        # For debugging: if our clock name is "OSCG", allow using the internal
        # oscillator. This is mostly useful for debugging.
        if clock_name == "OSCG":
            logging.warning("Using FPGA-internal oscillator for an approximately 62MHz.")
            logging.warning("USB communication won't work for f_OSC != 60MHz.")

            input_clock = Signal()
            m.submodules += Instance("OSCG", p_DIV=self.OSCG_DIV, o_OSC=input_clock)
            pll_params["CLKFB_DIV"] = 4
        else:
            input_clock = platform.request(clock_name).i

            divisor = 240e6 / clock_frequency
            if not divisor.is_integer():
                raise ValueError("Unsupported clock frequency {} MHz, must be an integer divisor of 240 MHz"
                                 .format(clock_frequency/1e6))

            pll_params["CLKFB_DIV"] = int(divisor)

        # Instantiate the ECP5 PLL.
        # These constants generated by Clarity Designer; which will
        # ideally be replaced by an open-source component.
        # (see https://github.com/SymbiFlow/prjtrellis/issues/34.)
        m.submodules.pll = Instance("EHXPLLL",

                # Clock in.
                i_CLKI=input_clock,

                # Generated clock outputs.
                o_CLKOP=self._clk_240MHz,
                o_CLKOS=self._clk_120MHz,
                o_CLKOS2=self._clk_60MHz,

                # Status.
                o_LOCK=self._pll_lock,

                # PLL parameters...
                p_PLLRST_ENA="DISABLED",
                p_INTFB_WAKE="DISABLED",
                p_STDBY_ENABLE="DISABLED",
                p_DPHASE_SOURCE="DISABLED",
                p_CLKOS3_FPHASE=0,
                p_CLKOS3_CPHASE=0,
                p_CLKOS2_FPHASE=0,
                p_CLKOS2_CPHASE=7,
                p_CLKOS_FPHASE=0,
                p_CLKOS_CPHASE=3,
                p_CLKOP_FPHASE=0,
                p_CLKOP_CPHASE=1,
                p_PLL_LOCK_MODE=0,
                p_CLKOS_TRIM_DELAY="0",
                p_CLKOS_TRIM_POL="FALLING",
                p_CLKOP_TRIM_DELAY="0",
                p_CLKOP_TRIM_POL="FALLING",
                p_OUTDIVIDER_MUXD="DIVD",
                p_CLKOS3_ENABLE="DISABLED",
                p_OUTDIVIDER_MUXC="DIVC",
                p_CLKOS2_ENABLE="ENABLED",
                p_OUTDIVIDER_MUXB="DIVB",
                p_CLKOS_ENABLE="ENABLED",
                p_OUTDIVIDER_MUXA="DIVA",
                p_CLKOP_ENABLE="ENABLED",
                p_CLKOS3_DIV=1,
                p_CLKOS2_DIV=8,
                p_CLKOS_DIV=4,
                p_CLKOP_DIV=2,
                p_CLKFB_DIV=pll_params["CLKFB_DIV"],
                p_CLKI_DIV=1,
                p_FEEDBK_PATH="CLKOP",

                # Internal feedback.
                i_CLKFB=self._clk_240MHz,

                # Control signals.
                i_RST=0,
                i_PHASESEL0=0,
                i_PHASESEL1=0,
                i_PHASEDIR=0,
                i_PHASESTEP=0,
                i_PHASELOADREG=0,
                i_STDBY=0,
                i_PLLWAKESYNC=0,

                # Output Enables.
                i_ENCLKOP=0,
                i_ENCLKOS=0,
                i_ENCLKOS2=0,
                i_ENCLKOS3=0,

                # Synthesis attributes.
                a_FREQUENCY_PIN_CLKI="60.000000",
                a_FREQUENCY_PIN_CLKOS2="60.000000",
                a_FREQUENCY_PIN_CLKOS="120.000000",
                a_FREQUENCY_PIN_CLKOP="240.000000",
                a_ICP_CURRENT="9",
                a_LPF_RESISTOR="8"
        )


        # Set up our global resets so the system is kept fully in reset until
        # our core PLL is fully stable. This prevents us from internally clock
        # glitching ourselves before our PLL is locked. :)
        m.d.comb += [
            ResetSignal("sync").eq(~self._pll_lock),
            ResetSignal("fast").eq(~self._pll_lock),
        ]


    def generate_usb_clock(self, m, platform):
        return self._clock_options[self.clock_frequencies['usb']]

    def generate_sync_clock(self, m, platform):
        return self._clock_options[self.clock_frequencies['sync']]

    def generate_fast_clock(self, m, platform):
        return self._clock_options[self.clock_frequencies['fast']]


    def stretch_sync_strobe_to_usb(self, m, strobe, output=None, allow_delay=False):
        """
        Helper that stretches a strobe from the `sync` domain to communicate with the `usn` domain.
        Works for any chosen frequency in which f(usb) < f(sync).
        """

        # TODO: replace with Amaranth's pulsesynchronizer?
        to_cycles = self.clock_frequencies['sync'] // self.clock_frequencies['usb']
        return stretch_strobe_signal(m, strobe, output=output, to_cycles=to_cycles, allow_delay=allow_delay)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Data header handling gateware.

This file currently contains very little logic; the actual transmission framing is handled at
the link layer; and the generation of our packets is handled by our endpoint.

"""

from amaranth import *
from usb_protocol.types.superspeed import HeaderPacketType

from ..link.header import HeaderQueue
from ..link.data   import DataHeaderPacket
from ...stream     import SuperSpeedStreamInterface
from ...request    import SetupPacket


class DataHeaderReceiver(Elaboratable):
    """ Gateware that handles received Data Header packets.

    Attributes
    -----------
    header_sink: HeaderQueue(), input stream
        Stream that brings up header packets for handling.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.header_sink   = HeaderQueue()


    def elaborate(self, platform):
        m = Module()

        # We handle Data Packets specially, passing their header data in conjunction with
        # the packets themselves; but the header packets are still handled like other header
        # packets, and must be explicitly consumed by the protocol layer.
        #
        # We'll consume all of them here, since we don't have any direct use for their data.
        new_packet = self.header_sink.valid
        is_for_us  = self.header_sink.get_type() == HeaderPacketType.DATA
        with m.If(new_packet & is_for_us):
            m.d.comb += self.header_sink.ready.eq(1)

        return m





#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test import LunaSSGatewareTestCase, ss_domain_test_case

from luna.gateware.usb.usb3.physical.scrambling import ScramblerLFSR

class ScramblerLFSRTest(LunaSSGatewareTestCase):
    FRAGMENT_UNDER_TEST = ScramblerLFSR

    @ss_domain_test_case
    def test_lfsr_stream(self):
        # From the table of 8-bit encoded values, [USB3.2, Appendix B.1].
        # We can continue this as long as we want to get more thorough testing,
        # but for now, this is probably enough.
        scrambled_sequence = [
            0x14c017ff, 0x8202e7b2, 0xa6286e72, 0x8dbf6dbe,   # Row 1 (0x00)
            0xe6a740be, 0xb2e2d32c, 0x2a770207, 0xe0be34cd,   # Row 2 (0x10)
            0xb1245da7, 0x22bda19b, 0xd31d45d4, 0xee76ead7    # Row 3 (0x20)
        ]

        yield self.dut.advance.eq(1)
        yield

        # Check that our LFSR produces each of our values in order.
        for index, value in enumerate(scrambled_sequence):
            self.assertEqual((yield self.dut.value), value, f"incorrect value at cycle {index}")
            yield




#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Standard, full-gateware control request handlers. """

from amaranth                 import *

from usb_protocol.types       import USBStandardRequests, USBRequestType
from usb_protocol.emitters    import DeviceDescriptorCollection

from ..application.request    import SuperSpeedRequestHandlerInterface
from ..application.descriptor import GetDescriptorHandler

from ...stream                import SuperSpeedStreamInterface
from ....utils                import falling_edge_detected


class StandardRequestHandler(Elaboratable):
    """ Pure-gateware USB3 setup request handler. Implements the standard requests required for enumeration. """

    def __init__(self, descriptors: DeviceDescriptorCollection):
        self.descriptors = descriptors

        #
        # I/O port
        #
        self.interface = SuperSpeedRequestHandlerInterface()



    def handle_register_write_request(self, m, new_value_signal, write_strobe, stall_condition=0):
        """ Fills in the current state with a request handler meant to set a register.

        Parameters
        ----------
        new_value_signal: Signal of any size
            The signal to receive the new value to be applied to the relevant register.
        write_strobe: Signal()
            The signal which will be pulsed when new_value_signal contains a update.
        stall_condition:
            If provided, if this condition is true, the request will be STALL'd instead of acknowledged.
            """

        # Provide an response to the STATUS stage.
        with m.If(self.interface.status_requested):

            # If our stall condition is met, stall; otherwise, send an ACK.
            with m.If(stall_condition):
                m.d.comb += self.interface.handshakes_out.send_stall.eq(1)
            with m.Else():
                m.d.comb += self.interface.handshakes_out.send_ack.eq(1)

            m.d.comb += [
                write_strobe      .eq(1),
                new_value_signal  .eq(self.interface.setup.value[0:7])
            ]

            # ... and then return to idle.
            m.next = 'IDLE'


    def handle_simple_data_request(self, m, data, *, length=1):
        """ Fills in a given current state with a request that returns a given short piece of data.

        For e.g. GET_CONFIGURATION and GET_STATUS requests. The relevant data must fit within a word.

        Parameters
        ----------
        tx_stream: StreamInterface
            The transmit stream to drive.
        data: Amaranth value, or equivalent, up to 32b
            The data to be transmitted.
        valid_mask: Amaranth value, or equivalent, up to 4b
            The valid mask for the data to be transmitted. Should be 0b0001, 0b0011, 0b0111, or 0b1111.
        """

        # Create a simple mapping of our valid bits to lengths.
        # Slightly clearer than the arithmetic version. :)
        valid_bits_for_length = [0b0000, 0b0001, 0b0011, 0b0111, 0b1111]

        # Provide our output stream with our simple word.
        tx_valid = Signal(4)
        m.d.comb += [
            self.interface.tx.valid   .eq(tx_valid),

            self.interface.tx.first   .eq(1),
            self.interface.tx.last    .eq(1),

            self.interface.tx.data    .eq(data),
            self.interface.tx_length  .eq(length)
        ]

        # When data is requested, start sending.
        with m.If(self.interface.data_requested):
            m.d.ss += tx_valid.eq(valid_bits_for_length[length])

        # Once our transmitter has accepted data, stop sending.
        with m.If(self.interface.tx.ready):
            m.d.ss += tx_valid.eq(0)

        # ACK our status stage, when appropriate.
        with m.If(self.interface.status_requested):
            m.d.comb += self.interface.handshakes_out.send_ack.eq(1)
            m.next = 'IDLE'



    def elaborate(self, platform):
        m = Module()
        interface = self.interface

        # Create convenience aliases for our interface components.
        setup               = interface.setup
        handshake_generator = interface.handshakes_out

        # For all of our handshakes, set our next-sequence-number to one; as we never receive
        # packets, and ACKs should be seq=1.
        m.d.comb += handshake_generator.next_sequence.eq(1)


        #
        # Submodules
        #

        # Handler for Get Descriptor requests; responds with our various fixed descriptors.
        m.submodules.get_descriptor = get_descriptor_handler = GetDescriptorHandler(self.descriptors,
            usb_domain  = "ss",
            stream_type = SuperSpeedStreamInterface
        )
        m.d.comb += [
            get_descriptor_handler.value  .eq(setup.value),
            get_descriptor_handler.length .eq(setup.length),
        ]


        ##
        ## Handlers.
        ##
        with m.If(setup.type == USBRequestType.STANDARD):
            with m.FSM(domain="ss"):

                # IDLE -- not handling any active request
                with m.State('IDLE'):

                    # If we've received a new setup packet, handle it.
                    with m.If(setup.received):

                        # Select which standard packet we're going to handler.
                        with m.Switch(setup.request):

                            with m.Case(USBStandardRequests.GET_STATUS):
                                m.next = 'GET_STATUS'
                            with m.Case(USBStandardRequests.SET_ADDRESS):
                                m.next = 'SET_ADDRESS'
                            with m.Case(USBStandardRequests.SET_CONFIGURATION):
                                m.next = 'SET_CONFIGURATION'
                            with m.Case(USBStandardRequests.GET_DESCRIPTOR):
                                m.next = 'GET_DESCRIPTOR'
                            with m.Case(USBStandardRequests.GET_CONFIGURATION):
                                m.next = 'GET_CONFIGURATION'
                            with m.Case(USBStandardRequests.SET_ISOCH_DELAY):
                                m.next = 'SET_ISOCH_DELAY'
                            with m.Case(USBStandardRequests.SET_SEL):
                                m.next = 'SET_SEL'
                            with m.Default():
                                m.next = 'UNHANDLED'


                # GET_STATUS -- Fetch the device's status.
                # For now, we'll always return '0'.
                with m.State('GET_STATUS'):
                    # TODO: handle reporting endpoint stall status
                    # TODO: copy the remote wakeup and bus-powered attributes from bmAttributes of the relevant descriptor?
                    self.handle_simple_data_request(m, 0, length=2)


                # SET_ADDRESS -- The host is trying to assign us an address.
                with m.State('SET_ADDRESS'):
                    self.handle_register_write_request(m, interface.new_address, interface.address_changed)


                # SET_CONFIGURATION -- The host is trying to select an active configuration.
                with m.State('SET_CONFIGURATION'):
                    # TODO: stall if we don't have a relevant configuration
                    self.handle_register_write_request(m, interface.new_config, interface.config_changed)


                # GET_DESCRIPTOR -- The host is asking for a USB descriptor -- for us to "self describe".
                with m.State('GET_DESCRIPTOR'):
                    m.d.comb += [
                        interface.tx                    .stream_eq(get_descriptor_handler.tx),
                        interface.tx_length             .eq(get_descriptor_handler.tx_length),

                        handshake_generator.send_stall  .eq(get_descriptor_handler.stall)
                    ]

                    # Respond to our data stage with a descriptor...
                    with m.If(interface.data_requested):
                        m.d.comb += get_descriptor_handler.start  .eq(1),

                    # ... and ACK our status stage.
                    with m.If(interface.status_requested):
                        m.d.comb += handshake_generator.send_ack.eq(1)
                        m.next = 'IDLE'


                # GET_CONFIGURATION -- The host is asking for the active configuration number.
                with m.State('GET_CONFIGURATION'):
                    self.handle_simple_data_request(m, interface.active_config)


                # SET_ISOCH_DELAY -- The host is trying to inform us of our isochronous delay.
                with m.State('SET_ISOCH_DELAY'):
                    # TODO: store this data aside once we support ISOCH
                    #self.handle_register_write_request(m, interface.new_config, interface.config_changed)

                    # ACK our status stage, when appropriate.
                    with m.If(self.interface.status_requested):
                        m.d.comb += self.interface.handshakes_out.send_ack.eq(1)
                        m.next = 'IDLE'


                # SET_SEL -- set our System Exit Latencies
                with m.State('SET_SEL'):
                    # TODO: use the actual latencies once we support USB3 power states

                    # ACK the data that's coming in, once we get it; but ignore it for now
                    data_received = falling_edge_detected(m, interface.rx.valid, domain="ss")
                    with m.If(data_received):
                        m.d.comb += self.interface.handshakes_out.send_ack.eq(1)


                    # ACK our status stage, when appropriate.
                    with m.If(self.interface.status_requested):
                        m.d.comb += self.interface.handshakes_out.send_ack.eq(1)
                        m.next = 'IDLE'


                # UNHANDLED -- we've received a request we're not prepared to handle
                with m.State('UNHANDLED'):

                    # When we next have an opportunity to stall, do so, and then return to idle.
                    data_received = falling_edge_detected(m, interface.rx.valid, domain="ss")
                    with m.If(interface.data_requested | interface.status_requested | data_received):
                        m.d.comb += handshake_generator.send_stall.eq(1)
                        m.next = 'IDLE'

        return m



# amaranth: UnusedElaboratable=no
#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Helpers for I/O interfacing. """

from amaranth import Record, Instance, Module, Signal, Cat
from amaranth.hdl.rec import DIR_FANIN, DIR_FANOUT


# FIXME: move this out of here and into an ECP5-specific set of functionality
def delay(m, signal, interval, *, out=None):
    """ Creates a delayed copy of a given I/O signal.

    Currently only works at the FPGA's I/O boundary, and only on ECP5s.

    Parameters:
        signal -- The signal to be delayed. Must be either an I/O
                  signal connected directly to a platform resource.
        delay  -- Delay, in arbitrary units. These units will vary
                  from unit to unit, but seem to be around 300ps on
                  most ECP5 FPGAs. On ECP5s, maxes out at 127.
        out    -- The signal to received the delayed signal; or
                  None ot have a signal created for you.

    Returns:
        delayed -- The delayed signal. Will be equivalent to 'out'
                   if provided; or a new signal otherwise.
    """

    # If we're not being passed our output signal, create one.
    if out is None:
        out = Signal.like(signal)

    # If we have more than one signal, call this function on each
    # of the subsignals.
    if len(signal) > 1:

        # If we have a vector of signals, but a integer delay,
        # convert that integer to a vector of same-valued delays.
        if isinstance(interval, int):
            interval = [interval] * len(signal)

        return Cat(delay(m, s, d, out=o) for s, d, o in zip(signal, interval, out))

    #
    # Base case: create a delayed version of the relevant signal.
    #
    m.submodules += Instance("DELAYG",
        i_A=signal,
        o_Z=out,
        p_DEL_VALUE=interval
    )

    return out



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" UTMI interfacing. """

from enum import IntEnum

from amaranth       import Elaboratable, Signal, Module
from amaranth.hdl.rec import Record, DIR_FANIN, DIR_FANOUT

from ..utils.bus    import OneHotMultiplexer

class UTMIOperatingMode:
    """ Enumeration that specifies the modes a UTMI transceiver can use. """

    NORMAL                    = 0
    NON_DRIVING               = 1

    RAW_DRIVE                 = 2
    DISABLE_BITSTUFF_AND_NRZI = 2
    CHIRP                     = 2

    NO_SYNC_OR_EOP            = 3


class UTMITerminationSelect:
    """ Enumeration that specifies meanings of the UTMI TermSelect bit. """

    HS_NORMAL    = 0
    HS_CHIRP     = 1
    LS_FS_NORMAL = 1


class UTMITransmitInterface(Record):
    """ Interface present on hardware that transmits onto a UTMI bus. """

    LAYOUT = [

        # Indicates when the data on tx_data is valid.
        ('valid', 1, DIR_FANOUT),

        # The data to be transmitted.
        ('data',  8, DIR_FANOUT),

        # Pulsed by the UTMI bus when the given data byte will be accepted
        # at the next clock edge.
        ('ready', 1, DIR_FANIN),
    ]

    def __init__(self):
        super().__init__(self.LAYOUT)


    def attach(self, utmi_bus):
        """ Returns a list of connection fragments connecting this interface to the provided bus.

        A typical usage might look like:
            m.d.comb += interface_object.attach(utmi_bus)
        """

        return [
            utmi_bus.tx_data   .eq(self.data),
            utmi_bus.tx_valid  .eq(self.valid),

            self.ready          .eq(utmi_bus.tx_ready),
        ]


class UTMIInterfaceMultiplexer(OneHotMultiplexer):
    """ Gateware that merges a collection of UTMITransmitInterfaces into a single interface.

    Assumes that only one transmitter will be communicating at once.

    I/O port:
        O*: output -- Our output interface; has all of the active busses merged together.
    """

    def __init__(self):
        super().__init__(
            interface_type=UTMITransmitInterface,
            mux_signals= ('data',),
            or_signals=  ('valid',),
            pass_signals=('ready',)
        )




class UTMIInterface(Record):
    """ UTMI+-standardized interface. Intended mostly as a simulation aid."""

    def __init__(self):
        super().__init__([
            # Core signals.
            ("rx_data",                     8),
            ("rx_active",                   1),
            ("rx_valid",                    1),

            ("tx_data",                     8),
            ("tx_valid",                    1),
            ("tx_ready",                    1),

            # Control signals.
            ('xcvr_select',                 2),
            ('term_select',                 1),
            ('op_mode',                     2),
            ('suspend',                     1),
            ('id_pullup',                   1),
            ('dm_pulldown',                 1),
            ('dp_pulldown',                 1),
            ('chrg_vbus',                   1),
            ('dischrg_vbus',                1),
            ('use_external_vbus_indicator', 1),

            # Event signals.
            ('line_state',                  2),
            ('vbus_valid',                  1),
            ('session_valid',               1),
            ('session_end',                 1),
            ('rx_error',                    1),
            ('host_disconnect',             1),
            ('id_digital',                  1)
        ])



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" USB3 link-layer abstraction."""

from amaranth import *

from ...stream          import USBRawSuperSpeedStream, SuperSpeedStreamArbiter, SuperSpeedStreamInterface
from ..physical.coding  import IDL

from .idle         import IdleHandshakeHandler
from .ltssm        import LTSSMController
from .header       import HeaderQueue, HeaderQueueArbiter
from .receiver     import HeaderPacketReceiver
from .transmitter  import PacketTransmitter
from .timers       import LinkMaintenanceTimers
from .ordered_sets import TSTransceiver
from .data         import DataPacketReceiver, DataPacketTransmitter, DataHeaderPacket


class USB3LinkLayer(Elaboratable):
    """ Abstraction encapsulating the USB3 link layer hardware.

    Performs the lower-level data manipulations associated with transporting USB3 packets
    from place to place.

    """

    def __init__(self, *, physical_layer, ss_clock_frequency=125e6):
        self._physical_layer  = physical_layer
        self._clock_frequency = ss_clock_frequency

        #
        # I/O port
        #

        # Header packet exchanges.
        self.header_sink               = HeaderQueue()
        self.header_source             = HeaderQueue()

        # Data packet exchange interface.
        self.data_source               = SuperSpeedStreamInterface()
        self.data_header_from_host     = DataHeaderPacket()
        self.data_source_complete      = Signal()
        self.data_source_invalid       = Signal()

        self.data_sink                 = SuperSpeedStreamInterface()
        self.data_sink_send_zlp        = Signal()
        self.data_sink_sequence_number = Signal(5)
        self.data_sink_endpoint_number = Signal(4)
        self.data_sink_length          = Signal(range(1024 + 1))
        self.data_sink_direction       = Signal()

        # Device state for header packets
        self.current_address           = Signal(7)

        # Status signals.
        self.trained                   = Signal()
        self.ready                     = Signal()
        self.in_reset                  = Signal()

        # Test and debug signals.
        self.disable_scrambling        = Signal()


    def elaborate(self, platform):
        m = Module()
        physical_layer = self._physical_layer

        # Mark ourselves as always consuming physical-layer packets.
        m.d.comb += physical_layer.source.ready.eq(1)

        #
        # Training Set Detectors/Emitters
        #
        training_set_source = USBRawSuperSpeedStream()

        m.submodules.ts = ts = TSTransceiver()
        m.d.comb += [
            # Note: we bring the physical layer's "raw" (non-descrambled) source to the TS detector,
            # as we'll still need to detect non-scrambled TS1s and TS2s if they arrive during normal
            # operation.
            ts.sink              .tap(physical_layer.raw_source),
            training_set_source  .stream_eq(ts.source)
        ]



        #
        # Idle handshake / logical idle detection.
        #
        m.submodules.idle = idle = IdleHandshakeHandler()
        m.d.comb += idle.sink.tap(physical_layer.source)


        #
        # U0 Maintenance Timers
        #
        m.submodules.timers = timers = LinkMaintenanceTimers(ss_clock_frequency=self._clock_frequency)


        #
        # Link Training and Status State Machine (LTSSM)
        #
        m.submodules.ltssm = ltssm = LTSSMController(ss_clock_frequency=self._clock_frequency)

        m.d.comb += [
            ltssm.phy_ready                      .eq(physical_layer.ready),

            # For now, we'll consider ourselves in USB reset iff we detect reset signaling.
            # This should be expanded; ideally to also consider e.g. loss of VBUS on some devices.
            ltssm.in_usb_reset                   .eq(physical_layer.lfps_reset_detected | ~physical_layer.vbus_present),

            # Link Partner Detection
            physical_layer.perform_rx_detection  .eq(ltssm.perform_rx_detection),
            ltssm.link_partner_detected          .eq(physical_layer.link_partner_detected),
            ltssm.no_link_partner_detected       .eq(physical_layer.no_link_partner_detected),

            # Pass down our link controls to the physical layer.
            physical_layer.tx_electrical_idle    .eq(ltssm.tx_electrical_idle),
            physical_layer.engage_terminations   .eq(ltssm.engage_terminations),
            physical_layer.invert_rx_polarity    .eq(ltssm.invert_rx_polarity),
            physical_layer.train_equalizer       .eq(ltssm.train_equalizer),

            # LFPS control.
            ltssm.lfps_polling_detected          .eq(physical_layer.lfps_polling_detected),
            physical_layer.send_lfps_polling     .eq(ltssm.send_lfps_polling),
            ltssm.lfps_cycles_sent               .eq(physical_layer.lfps_cycles_sent),

            # Training set detectors
            ltssm.tseq_detected                  .eq(ts.tseq_detected),
            ltssm.ts1_detected                   .eq(ts.ts1_detected),
            ltssm.inverted_ts1_detected          .eq(ts.inverted_ts1_detected),
            ltssm.ts2_detected                   .eq(ts.ts2_detected),
            ltssm.hot_reset_requested            .eq(ts.hot_reset_requested),
            ltssm.loopback_requested             .eq(ts.loopback_requested),
            ltssm.no_scrambling_requested        .eq(ts.no_scrambling_requested),

            # Training set emitters
            ts.send_tseq_burst                   .eq(ltssm.send_tseq_burst),
            ts.send_ts1_burst                    .eq(ltssm.send_ts1_burst),
            ts.send_ts2_burst                    .eq(ltssm.send_ts2_burst),
            ts.request_hot_reset                 .eq(ltssm.request_hot_reset),
            ts.request_no_scrambling             .eq(ltssm.request_no_scrambling),
            ltssm.ts_burst_complete              .eq(ts.burst_complete),

            # Scrambling control.
            physical_layer.enable_scrambling     .eq(ltssm.enable_scrambling),

            # Idle detection.
            idle.enable                          .eq(ltssm.perform_idle_handshake),
            ltssm.idle_handshake_complete        .eq(idle.idle_handshake_complete),

            # Link maintainance.
            timers.enable                        .eq(ltssm.link_ready),

            # Status signaling.
            self.trained                         .eq(ltssm.link_ready),
            self.in_reset                        .eq(ltssm.request_hot_reset | ltssm.in_usb_reset),

            # Test and debug.
            ltssm.disable_scrambling             .eq(self.disable_scrambling),
        ]


        #
        # Packet transmission path.
        # Accepts packets from the protocol and link layers, and transmits them.
        #

        # Transmit header multiplexer.
        m.submodules.hp_mux = hp_mux = HeaderQueueArbiter()
        hp_mux.add_producer(self.header_sink)

        # Core transmitter.
        m.submodules.transmitter = transmitter = PacketTransmitter()
        m.d.comb += [
            transmitter.sink                .tap(physical_layer.source),
            transmitter.enable              .eq(ltssm.link_ready),
            transmitter.usb_reset           .eq(self.in_reset),

            transmitter.queue               .header_eq(hp_mux.source),

            # Link state management handling.
            timers.link_command_received  .eq(transmitter.link_command_received),
            self.ready                    .eq(transmitter.bringup_complete),
        ]



        #
        # Header Packet Rx Path.
        # Receives header packets and forwards them up to the protocol layer.
        #
        m.submodules.header_rx = header_rx = HeaderPacketReceiver()
        m.d.comb += [
            header_rx.sink                   .tap(physical_layer.source),
            header_rx.enable                 .eq(ltssm.link_ready),
            header_rx.usb_reset              .eq(self.in_reset),

            # Bring our header packet interface to the protocol layer.
            self.header_source               .header_eq(header_rx.queue),

            # Keepalive handling.
            timers.link_command_transmitted  .eq(header_rx.source.valid),
            header_rx.keepalive_required     .eq(timers.schedule_keepalive),
            timers.packet_received           .eq(header_rx.packet_received),

            # Transmitter event path.
            header_rx.retry_required         .eq(transmitter.retry_required),
            transmitter.lrty_pending         .eq(header_rx.lrty_pending),
            header_rx.retry_received         .eq(transmitter.retry_received),

            # For now, we'll reject all forms of power management by sending a REJECT
            # whenever we receive an LGO (Link Go-to) request.
            header_rx.reject_power_state     .eq(transmitter.lgo_received),
        ]


        #
        # Link Recovery Control
        #
        m.d.comb += ltssm.trigger_link_recovery.eq(
            timers.transition_to_recovery |
            header_rx.recovery_required   |
            transmitter.recovery_required
        )

        #
        # Data packet handlers.
        #

        # Receiver.
        m.submodules.data_rx = data_rx = DataPacketReceiver()
        m.d.comb += [
            data_rx.sink                .tap(physical_layer.source),

            # Data interface to Protocol layer.
            self.data_source            .stream_eq(data_rx.source),
            self.data_header_from_host  .eq(data_rx.header),
            self.data_source_complete   .eq(data_rx.packet_good),
            self.data_source_invalid    .eq(data_rx.packet_bad),
        ]

        # Transmitter.
        m.submodules.data_tx = data_tx = DataPacketTransmitter()
        hp_mux.add_producer(data_tx.header_source)

        m.d.comb += [
            transmitter.data_sink    .stream_eq(data_tx.data_source),

            # Device state information.
            data_tx.address          .eq(self.current_address),

            # Data interface from Protocol layer.
            data_tx.data_sink        .stream_eq(self.data_sink),
            data_tx.send_zlp         .eq(self.data_sink_send_zlp),
            data_tx.sequence_number  .eq(self.data_sink_sequence_number),
            data_tx.endpoint_number  .eq(self.data_sink_endpoint_number),
            data_tx.data_length      .eq(self.data_sink_length),
            data_tx.direction        .eq(self.data_sink_direction)
        ]


        #
        # Transmit stream arbiter.
        #
        m.submodules.stream_arbiter = arbiter = SuperSpeedStreamArbiter()

        # Add each of our streams to our arbiter, from highest to lowest priority.
        arbiter.add_stream(training_set_source)
        arbiter.add_stream(header_rx.source)
        arbiter.add_stream(transmitter.source)

        # If we're idle, send logical idle.
        with m.If(arbiter.idle):
            m.d.comb += [
                # Drive our idle stream with our IDL value (0x00)...
                physical_layer.sink.valid    .eq(1),
                physical_layer.sink.data     .eq(IDL.value),
                physical_layer.sink.ctrl     .eq(IDL.ctrl),

                # Let the physical layer know it can insert CTC skips whenever data is being accepted
                # from our logical idle stream.
                physical_layer.can_send_skp  .eq(1)
            ]

        # Otherwise, output our stream data.
        with m.Else():
            m.d.comb += physical_layer.sink.stream_eq(arbiter.source)



        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test import LunaSSGatewareTestCase, ss_domain_test_case

from luna.gateware.usb.usb3.physical.ctc import CTCSkipRemover

class CTCSkipRemoverTest(LunaSSGatewareTestCase):
    FRAGMENT_UNDER_TEST = CTCSkipRemover

    def initialize_signals(self):
        # Set up our streams to always ferry data in and out, where possible.
        yield self.dut.sink.valid.eq(1)
        yield self.dut.source.ready.eq(1)


    def provide_input(self, data, ctrl):
        yield self.dut.sink.data.eq(data)
        yield self.dut.sink.ctrl.eq(ctrl)
        yield


    @ss_domain_test_case
    def test_dual_skip_removal(self):
        source = self.dut.source

        # When we add data into the buffer...
        yield from self.provide_input(0xAABBCCDD, 0b0000)

        # ... we should see our line go valid only after four bytes are collected.
        self.assertEqual((yield source.valid), 0)
        yield from self.provide_input(0x71BA3C3C, 0b0011)

        # Once it does go high, it should be accompanied by valid input data.
        self.assertEqual((yield source.valid), 1)
        self.assertEqual((yield source.data), 0xAABBCCDD)
        self.assertEqual((yield source.ctrl), 0)
        yield from self.provide_input(0x11223344, 0b1100)

        # If data with SKPs were provided, our output should be invalid, until we
        # receive enough bytes to have four non-skip bytes.
        self.assertEqual((yield source.valid), 0)

        # Once we do, we should see a copy of our data without the SKPs included.
        yield
        self.assertEqual((yield source.data), 0x334471BA)
        self.assertEqual((yield source.ctrl), 0)
        yield
        self.assertEqual((yield source.data), 0x33441122)
        self.assertEqual((yield source.ctrl), 0b11)


    @ss_domain_test_case
    def test_shifted_dual_skip_removal(self):
        source = self.dut.source

        # When we add data into the buffer...
        yield from self.provide_input(0xAABBCCDD, 0b0000)

        # ... we should see our line go valid only after four bytes are collected.
        self.assertEqual((yield source.valid), 0)
        yield from self.provide_input(0x713C3CBA, 0b0110)

        # Once it does go high, it should be accompanied by valid input data.
        self.assertEqual((yield source.valid), 1)
        self.assertEqual((yield source.data), 0xAABBCCDD)
        self.assertEqual((yield source.ctrl), 0)
        yield from self.provide_input(0x113C3C44, 0b0110)

        # If data with SKPs were provided, our output should be invalid, until we
        # receive enough bytes to have four non-skip bytes.
        self.assertEqual((yield source.valid), 0)

        # Once we do, we should see a copy of our data without the SKPs included.
        yield from self.provide_input(0x55667788, 0b0000)
        self.assertEqual((yield source.data), 0x114471BA)
        self.assertEqual((yield source.ctrl), 0)
        yield
        self.assertEqual((yield source.data), 0x55667788)
        self.assertEqual((yield source.ctrl), 0)


    @ss_domain_test_case
    def test_single_skip_removal(self):
        source = self.dut.source

        # When we add data into the buffer...
        yield from self.provide_input(0xAABBCCDD, 0b0000)

        # ... we should see our line go valid only after four bytes are collected.
        self.assertEqual((yield source.valid), 0)
        yield from self.provide_input(0x3C556677, 0b1000)

        # Once it does go high, it should be accompanied by valid input data.
        self.assertEqual((yield source.valid), 1)
        self.assertEqual((yield source.data), 0xAABBCCDD)
        self.assertEqual((yield source.ctrl), 0)
        yield from self.provide_input(0x11223344, 0b1100)

        # If data with SKPs were provided, our output should be invalid, until we
        # receive enough bytes to have four non-skip bytes.
        self.assertEqual((yield source.valid), 0)

        # Once we do, we should see a copy of our data without the SKPs included.
        yield
        self.assertEqual((yield source.data), 0x44556677)
        self.assertEqual((yield source.ctrl), 0)
        yield
        self.assertEqual((yield source.data), 0x44112233)
        self.assertEqual((yield source.ctrl), 0b110)


    @ss_domain_test_case
    def test_cycle_spread_skip_removal(self):
        source = self.dut.source

        # When we add data into the buffer...
        yield from self.provide_input(0xAABBCCDD, 0b0000)

        # ... we should see our line go valid only after four bytes are collected.
        self.assertEqual((yield source.valid), 0)
        yield from self.provide_input(0x3C556677, 0b1000)

        # Once it does go high, it should be accompanied by valid input data.
        self.assertEqual((yield source.valid), 1)
        self.assertEqual((yield source.data), 0xAABBCCDD)
        self.assertEqual((yield source.ctrl), 0)
        yield from self.provide_input(0x1122333C, 0b0001)

        # If data with SKPs were provided, our output should be invalid, until we
        # receive enough bytes to have four non-skip bytes.
        self.assertEqual((yield source.valid), 0)

        # Once we do, we should see a copy of our data without the SKPs included.
        yield from self.provide_input(0x44556677, 0b0000)
        self.assertEqual((yield source.data), 0x33556677)
        self.assertEqual((yield source.ctrl), 0)
        yield
        self.assertEqual((yield source.data), 0x66771122)
        self.assertEqual((yield source.ctrl), 0b0)

